/******************************************************************
 FILE:		Led.c
 CONTENTS:	Definitions for structures, methods, and actions of the 
		LedButton widget.
 AUTHOR:	Paul D. Johnston
 HISTORY:
 Date		Action
 ---------	------------------------------------
 10/22/91	Created.

********************************************************************/

#include <stdio.h>

/* Xlib includes */
#include <X11/Xlib.h>

/* Xt includes */
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>


/* Widget includes */
#include <X11/Xc/LedP.h>	/* (includes Led.h also) */


/* Macro redefinition for offset. */
#define offset(field) XtOffset(LedButtonWidget, field)


/* Declare widget methods */
static void ClassInitialize();
static void Initialize();
static void Redisplay();
static void Resize();
static void Destroy();
static XtGeometryResult QueryGeometry();
static Boolean SetValues();


/* Declare functions and variables private to this widget */
void Draw_value();
void Get_value();

/* Define the widget's resource list */
static XtResource resources[] =
{
  {
    XcNledVisible,
    XtCBoolean,
    XtRBoolean,
    sizeof(Boolean),
    offset(ledButton.led_visible),
    XtRString,
    "True"
  },
  {
    XcNstateVisible,
    XtCBoolean,
    XtRBoolean,
    sizeof(Boolean),
    offset(ledButton.state_visible),
    XtRString,
    "False"
  },
  {
    XcNstate,
    XcCState,
    XtRString,
    sizeof(String),
    offset(ledButton.state),
    XtRString,
    " "
  },
  {
    XcNupdateCallback,
    XtCCallback,
    XtRCallback,
    sizeof(caddr_t),
    offset(ledButton.callback),
    XtRCallback,
    NULL
  },
  {
    XcNstateForeground,
    XtCColor,
    XtRPixel,
    sizeof(Pixel),
    offset(ledButton.state_fg_pixel),
    XtRString,
    XtDefaultForeground
  },
  {
    XcNstateBackground,
    XtCColor,
    XtRPixel,
    sizeof(Pixel),
    offset(ledButton.state_bg_pixel),
    XtRString,
    "white"
  },
  {
    XcNledColor,
    XtCColor,
    XtRPixel,
    sizeof(Pixel),
    offset(ledButton.led_color),
    XtRString,
    "white"
  },
  {
    XcNinterval,
    XcCInterval,
    XtRInt,
    sizeof(int),
    offset(ledButton.interval),
    XtRImmediate,
    (XtPointer)0
  },
};



/* Widget Class Record initialization */
LedButtonClassRec ledButtonClassRec =
{
  {
  /* core_class part */
    (WidgetClass) &controlClassRec,		/* superclass */
    "LedButton",				/* class_name */
    sizeof(LedButtonRec),			/* widget_size */
    NULL,					/* class_initialize */
    NULL,					/* class_part_initialize */
    FALSE,					/* class_inited */
    Initialize,					/* initialize */
    NULL,					/* initialize_hook */
    XtInheritRealize,				/* realize */
    NULL,					/* actions */
    0,						/* num_actions */
    resources,					/* resources */
    XtNumber(resources),			/* num_resources */
    NULLQUARK,					/* xrm_class */
    TRUE,					/* compress_motion */
    TRUE,					/* compress_exposure */
    TRUE,					/* compress_enterleave */
    FALSE,					/* visible_interest */
    Destroy,					/* destroy */
    Resize,					/* resize */
    Redisplay,					/* expose */
    SetValues,					/* set_values */
    NULL,					/* set_values_hook */
    XtInheritSetValuesAlmost,			/* set_values_almost */
    NULL,					/* get_values_hook */
    NULL,					/* accept_focus */
    XtVersion,					/* version */
    NULL,					/* callback_private */
    NULL,					/* tm_table */ 
    QueryGeometry,				/* query_geometry */
    XtInheritDisplayAccelerator,		/* display_accelerator */
    NULL					/* extension */
  },
  {
  /* Control class part */
    0,						/* dummy_field */
  },
  {
  /* Led class part */
    0,						/* dummy_field */
  },
};

WidgetClass xcLedButtonWidgetClass = (WidgetClass)&ledButtonClassRec;


/* Widget method function definitions */

/*******************************************************************
 NAME:		Initialize.		
 DESCRIPTION:
   This is the initialize method for the LedButton widget.  It 
validates user-modifiable instance resources and initializes private 
widget variables and structures.  This function also creates any server 
resources (i.e., GCs, fonts, Pixmaps, etc.) used by this widget.  This
method is called by Xt when the application calls XtCreateWidget().

*******************************************************************/

static void Initialize(request, new)
LedButtonWidget request, new;
{
/* Local variables */
Display *display = XtDisplay(new);
int scr = DefaultScreen(XtDisplay(new));


DPRINTF(("LedButton: executing Initialize...\n"));

/*
 * Validate public instance variable settings.
 */
   
   if (new->ledButton.interval > 0)
      new->ledButton.interval_id = XtAddTimeOut(new->ledButton.interval, 
							Get_value, new);

   if ((new->ledButton.led_visible != True) && 
       (new->ledButton.led_visible != False))
   {
      XtWarning("LedButton: invalid ledVisible setting");
      new->ledButton.led_visible = True;
   }

   if ((new->ledButton.state_visible != True) && 
       (new->ledButton.state_visible != False))
   {
      XtWarning("LedButton: invalid stateVisible setting");
      new->ledButton.state_visible = False;
   }

   if ((new->ledButton.led_visible == False) && 
       (new->ledButton.state_visible == False))
   {
      XtWarning("LedButton: ledVisible/stateVisible mismatch");
      new->ledButton.led_visible = True;
   }


/* Initialize the LedButton width and height. */
   if (new->core.width < MIN_LB_WIDTH)
      new->core.width = MIN_LB_WIDTH; 
   if (new->core.height < MIN_LB_HEIGHT)
      new->core.height = MIN_LB_HEIGHT; 
   
/* Initialize private instance variables.  */
   new->ledButton.armed = False;
   
/* Set the initial geometry of the LedButton elements. */
   Resize(new);


DPRINTF(("LedButton: done Initialize\n"));

}  /* end of Initialize */





/*******************************************************************
 NAME:		Redisplay.	
 DESCRIPTION:
   This function is the LedButton's Expose method.  It redraws the
LedButton's 3D rectangle background, and label string. All drawing 
takes place within the widget's window (no need for an off-screen pixmap).

*******************************************************************/

static void Redisplay(w, event)
LedButtonWidget w;
XExposeEvent *event;
{

/*
 * Check to see whether or not the widget's window is mapped.  You can't
 * draw into a window that is not mapped.  Realizing a widget doesn't 
 * mean its mapped, but this call will work for most Window Managers.
 */
   if (!XtIsRealized(w))
      return;

DPRINTF(("LedButton: executing Redisplay\n"));

/* Draw the background button */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	0, 0, w->core.width, w->core.height, RAISED);

   XSetClipRectangles(XtDisplay(w), w->control.gc, 0, 0, 
  			&(w->ledButton.face), 1, Unsorted); 

/* Draw label and label background color */
   XSetForeground(XtDisplay(w), w->control.gc,
                        w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), w->control.gc,
        w->ledButton.lbl.x, w->ledButton.lbl.y,
        w->control.label, strlen(w->control.label));

/* Draw the State Box */
   if (w->ledButton.state_visible == True)
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->ledButton.state_box.x - w->control.shade_depth, 
	w->ledButton.state_box.y - w->control.shade_depth,
	w->ledButton.state_box.width + (2 * w->control.shade_depth),  
	w->ledButton.state_box.height + (2 * w->control.shade_depth), 
	DEPRESSED);

/* Draw the LED box */
   if (w->ledButton.led_visible == True)
   {
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->ledButton.led_face.x,  w->ledButton.led_face.y,
	w->ledButton.led_face.width,  
	w->ledButton.led_face.height, 
	RAISED);
      VarRect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->ledButton.led_face.x + w->control.shade_depth,  
	w->ledButton.led_face.y + w->control.shade_depth,
	w->ledButton.led_face.width - 2*w->control.shade_depth,  
	w->ledButton.led_face.height - 2*w->control.shade_depth, 
	DEPRESSED, w->ledButton.inner_shade_depth);
   }

/* Draw the new LED color/state string. */
   Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);


DPRINTF(("LedButton: done Redisplay\n"));

}  /* end of Redisplay */





/*******************************************************************
 NAME:		SetValues.
 DESCRIPTION:
   This is the set_values method for this widget. It validates resource
settings set with XtSetValues. If a resource is changed that would
require re-drawing the widget, return True.

*******************************************************************/

static Boolean SetValues(cur, req, new)
LedButtonWidget cur, req, new;
{
/* Local variables */
Boolean do_redisplay = False;

DPRINTF(("LedButton: executing SetValues for %ld\n", cur));

/* Validate new resource settings. */

/* Check to see if the ledVisible or stateVisible resources have changed */
   if (new->ledButton.led_visible != cur->ledButton.led_visible)
   {
      do_redisplay = True;
      if ((new->ledButton.led_visible != True) && 
          (new->ledButton.led_visible != False))
      {
         XtWarning("LedButton: invalid ledVisible setting");
         new->ledButton.led_visible = True;
      }
   }

   if (new->ledButton.state_visible != cur->ledButton.state_visible)
   {
      do_redisplay = True;
      if ((new->ledButton.state_visible != True) && 
          (new->ledButton.state_visible != False))
      {
         XtWarning("LedButton: invalid stateVisible setting");
         new->ledButton.state_visible = False;
      }
   }

   if ((new->ledButton.led_visible == False) && 
       (new->ledButton.state_visible == False))
   {
      XtWarning("LedButton: ledVisible/stateVisible mismatch");
      new->ledButton.led_visible = True;
   }

DPRINTF(("LedButton: done SetValues\n"));
   return do_redisplay;


}  /* end of SetValues */




/*******************************************************************
 NAME:		Resize.
 DESCRIPTION:
   This is the resize method of the LedButton widget. It resizes the
LedButton's graphics based on the new width and height of the widget's
window.  

*******************************************************************/

static void Resize(w)
LedButtonWidget w;
{
  Pixel strwidth2;
  Pixel facewidth4;

DPRINTF(("LedButton: executing Resize\n"));

/* Set the widget's new width and height. */
   w->ledButton.face.x = w->ledButton.face.y = w->control.shade_depth;
   w->ledButton.face.width = w->core.width - (2*w->control.shade_depth)+1;
   w->ledButton.face.height = w->core.height - (2*w->control.shade_depth)+1;

/* Set the LED's new width and height. */
   w->ledButton.led_face.x = w->ledButton.led_face.y = 
			w->ledButton.face.x + w->control.shade_depth;
   w->ledButton.led_face.width = w->ledButton.face.width / 2  -
						(2*w->control.shade_depth);
   w->ledButton.led_face.height = w->ledButton.face.height -
						(2*w->control.shade_depth);

/* set the LED's inner shade_depth */
   w->ledButton.inner_shade_depth = (int)(w->control.shade_depth / 4);


/* Establish the new State Box geometry. */
   w->ledButton.state_box.y = w->ledButton.face.y +
		(w->ledButton.face.height / 2 + 0.5) + w->control.shade_depth;
   w->ledButton.state_box.height = (w->ledButton.face.height / 2 + 0.5) -
				 	(2 * w->control.shade_depth);
   if (w->ledButton.led_visible == True)
   {
      w->ledButton.state_box.x = w->ledButton.face.x + 
		(w->ledButton.face.width / 2 + 0.5) + w->control.shade_depth;
      w->ledButton.state_box.width = (w->ledButton.face.width / 2 + 0.5) -
				 	(2 * w->control.shade_depth);
   }
   else
   {
      w->ledButton.state_box.width = ((3 * w->ledButton.face.width) / 4 +0.5)
					- (2 * w->control.shade_depth);
      w->ledButton.state_box.x = w->ledButton.face.x + 
		(w->ledButton.face.width / 2 + 0.5) -
		(w->ledButton.state_box.width / 2 + 0.5) +
						w->control.shade_depth;
   }

/* Calculate new State String center position */
   w->ledButton.st.x = w->ledButton.state_box.x +
	 	((w->ledButton.state_box.width - 
         	  XTextWidth(w->control.font, w->ledButton.state, 
		    	    strlen(w->ledButton.state))) / 2);
   w->ledButton.st.y = w->ledButton.state_box.y +
		(w->ledButton.state_box.height / 2 + 0.5) + 
		((w->control.font->ascent + w->control.font->descent) / 2 +0.5)
				- w->control.font->descent;

/* Set the Label location based on whether or not the State Box is visible. */
   if (w->ledButton.led_visible)
      w->ledButton.lbl.x = w->ledButton.state_box.x - w->control.shade_depth;
   else
      w->ledButton.lbl.x = w->ledButton.face.x +
		(w->ledButton.face.width / 2) -
        	(XTextWidth(w->control.font, w->control.label, 
		    	    strlen(w->control.label)) / 2);
      
   if (w->ledButton.state_visible)
   {
      w->ledButton.lbl.y = w->ledButton.face.y + 
		((w->ledButton.state_box.y - w->control.shade_depth - 
		  			w->ledButton.face.y) / 2 + 0.5) +
		((w->control.font->ascent + w->control.font->descent) / 2)
						- w->control.font->descent; 
   }
   else
   {
      w->ledButton.lbl.y = w->ledButton.face.y +
		(w->ledButton.face.height / 2 + 0.5) + 
		((w->control.font->ascent + w->control.font->descent) / 2)
						- w->control.font->descent; 
   }

DPRINTF(("LedButton: done Resize\n"));

}  /* end of Resize */




/*******************************************************************
 NAME:		QueryGeometry.		
 DESCRIPTION:
   This function is the widget's query_geometry method.  It simply
checks the proposed size and returns the appropriate value based on
the proposed size.  If the proposed size is greater than the maximum
appropriate size for this widget, QueryGeometry returns the recommended
size.

*******************************************************************/

static XtGeometryResult QueryGeometry(w, proposed, answer)
LedButtonWidget w;
XtWidgetGeometry *proposed, *answer;
{
/* Set the request mode mask for the returned answer. */
   answer->request_mode = CWWidth | CWHeight;

/* Set the recommended size. */
   answer->width = (w->core.width > MAX_LB_WIDTH)
	? MAX_LB_WIDTH : w->core.width;
   answer->height = (w->core.height > MAX_LB_HEIGHT)
	? MAX_LB_HEIGHT : w->core.height;

/* 
 * Check the proposed dimensions. If the proposed size is larger than
 * appropriate, return the recommended size.
 */
   if (((proposed->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight))
	&& proposed->width == answer->width 
	&& proposed->height == answer->height)
      return XtGeometryYes;
   else if (answer->width == w->core.width && answer->height == w->core.height)
      return XtGeometryNo;
   else
      return XtGeometryAlmost;

}  /* end of QueryGeometry */



/*******************************************************************
 NAME:		Destroy.
 DESCRIPTION:
   This function is the widget's destroy method.  It simply releases
any server resources acquired during the life of the widget.

*******************************************************************/

static void Destroy(w)
LedButtonWidget w;
{

   if (w->ledButton.interval > 0)
      XtRemoveTimeOut (w->ledButton.interval_id);

}  /* end of Destroy */




/* Widget action functions. */

/*******************************************************************
 NAME:		Draw_value.		
 DESCRIPTION:
   This function redraws the value string in the State Box.

*******************************************************************/

void Draw_value(

	w,

	display,

	drawable,

	gc)

LedButtonWidget w;
Display *display;
Drawable drawable;
GC gc;
{
/* Local variables */
char temp[20];
XRectangle val_clip;

/* Draw the new LED color. */
   if (w->ledButton.led_visible == True)
   {
      XSetForeground(display, gc, w->ledButton.led_color);  
      XFillRectangle(display, drawable, gc,
         w->ledButton.led_face.x + w->control.shade_depth + 
	    w->ledButton.inner_shade_depth, 
         w->ledButton.led_face.y + w->control.shade_depth +
	    w->ledButton.inner_shade_depth,  
         w->ledButton.led_face.width - (2 * w->control.shade_depth) -
	    (2 * w->ledButton.inner_shade_depth),  
         w->ledButton.led_face.height - (2 * w->control.shade_depth) -
	    (2 * w->ledButton.inner_shade_depth));  
   }

  if (w->ledButton.state_visible == True)
  {
	/* Clear the State Box by re-filling it with its background color. */
     XSetForeground(display, gc, w->ledButton.state_bg_pixel);  
     XFillRectangle(display, drawable, gc,
  	w->ledButton.state_box.x, w->ledButton.state_box.y, 
  	w->ledButton.state_box.width + 1, w->ledButton.state_box.height + 1); 

	/*
 	 * Now draw the value string in its foreground color, clipped by the
 	 * State Box.
 	 */
     XSetForeground(display, gc, w->ledButton.state_fg_pixel);  
     XSetClipRectangles(display, gc, 0, 0, 
  	&(w->ledButton.state_box), 1, Unsorted); 
     XDrawString(display, drawable, gc,
    	w->ledButton.st.x, w->ledButton.st.y, w->ledButton.state, 
	strlen(w->ledButton.state)); 
  }

/* Reset the clip_mask to no clipping. */
   XSetClipMask(display, gc, None);

}  /* end of Draw_value */




/*******************************************************************
Name:		Get_value 
DESCRIPTION:
	This function gets LED color and state string information
********************************************************************/
void Get_value(w, id)

LedButtonWidget w;
XtIntervalId id;
{
  /* local variables */  
  XcLedInfoRec ledinfo;
 
   /* register another timeout function call */
   w->ledButton.interval_id = XtAddTimeOut(w->ledButton.interval, 
							Get_value, w);
  
   if (w->ledButton.callback != NULL)
      XtCallCallbacks(w, XcNupdateCallback, &ledinfo);

/* redraw LED color and state string */

  XcLedSetState(w, ledinfo.color, ledinfo.str);

} /* end of Get_value */


/*******************************************************************
 NAME:		XcLedSetState		
 DESCRIPTION:
	This function sets LED color and state string.
*******************************************************************/

void XcLedSetState(
	w,
	colorname,
	str)

LedButtonWidget w; 
char *colorname;
char *str;
{
/* local variables */
  Display *dpy = XtDisplay(w);
  int scr  = DefaultScreen(dpy);
  Colormap cmap = DefaultColormap(dpy, scr);
  XColor color, ignore;
  char  errstring[40];

/* allocate the named color */
   if (w->ledButton.led_visible == True)
   {
      if (XParseColor(dpy, cmap, colorname, &color))
      {
        if (XAllocColor(dpy, cmap, &color))
        {
           w->ledButton.led_color = color.pixel;
        }
        else
        {
           sprintf(errstring, "LedButton: Couldn't allocate %s", colorname);
           XtWarning(errstring);
           w->ledButton.led_color = BlackPixel(dpy, scr);
        } 
      }
      else
      {
        sprintf(errstring, "LedButton: Couldn't pixel for %s", colorname);
        XtWarning(errstring);
        w->ledButton.led_color = BlackPixel(dpy, scr);
      } 
   }
   
   if (str != NULL)
      strcpy(w->ledButton.state, str);

/* calculate new state string center position */
   w->ledButton.st.x = w->ledButton.state_box.x +
	 	((w->ledButton.state_box.width - 
         	  XTextWidth(w->control.font, w->ledButton.state, 
		    	    strlen(w->ledButton.state))) / 2);

/* redraw LED color and state string */
  if (XtIsRealized(w))
     Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);

}  /* end of XcLedSetState */



/* end of Led.c */

