/*******************************************************************
 FILE:		SimpleB.c
 CONTENTS:	Definitions for structures, methods, and actions of the 
		SimpleButton widget.
 AUTHOR:	Paul D. Johnston
 HISTORY:
 Date		Action
 ---------	------------------------------------
 5/23/92	Changed the widget class name so that it is preceded
		by 'xc' with the first major word capitalized.
 10/22/91	Created.

********************************************************************/

#include <stdio.h>

/* Xlib includes */
#include <X11/Xlib.h>

/* Xt includes */
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#ifdef X11R3
#include <X11/Xmu.h>
#else
#include <X11/Xmu/Xmu.h>
#endif


/* Widget includes */
#include <X11/Xc/SimpleBP.h>	/* (includes SimpleB.h also) */


/* Macro redefinition for offset. */
#define offset(field) XtOffset(SimpleButtonWidget, field)


/* Declare widget methods */
static void ClassInitialize();
static void Initialize();
static void Redisplay();
static void Resize();
static XtGeometryResult QueryGeometry();
static Boolean SetValues();

/* Declare widget actions */
static void Do_button();	/* Handles all ButtonDown and ButtonUp 
				 * events in the SimpleButton widget.
				 */
static void Do_enter_leave();	/* Handles all EnterNotify and LeaveNotify
				 * events in the SimpleButton widget.
				 */



/* Declare functions and variables private to this widget */
void Draw_button();



/* Define the widget's resource list */
static XtResource resources[] =
{
  {
    XcNactivateCallback,
    XtCCallback,
    XtRCallback,
    sizeof(caddr_t),
    offset(simpleButton.callback),
    XtRCallback,
    NULL
  },

};


/* Widget translation and actions tables. */
static XtActionsRec actions[] =
{
   { "Do_button", Do_button },
   { "Do_enter_leave", Do_enter_leave }
};

static char defaultTranslations[] = "<Btn1Down>:  Do_button() \n\
				     <Enter>:  Do_enter_leave() \n\
				     <Leave>:  Do_enter_leave() \n\
				     <Btn1Up>:  Do_button() ";


/* Widget Class Record initialization */
SimpleButtonClassRec simpleButtonClassRec =
{
  {
  /* core_class part */
    (WidgetClass) &controlClassRec,		/* superclass */
    "SimpleButton",				/* class_name */
    sizeof(SimpleButtonRec),			/* widget_size */
    NULL,					/* class_initialize */
    NULL,					/* class_part_initialize */
    FALSE,					/* class_inited */
    Initialize,					/* initialize */
    NULL,					/* initialize_hook */
    XtInheritRealize,				/* realize */
    actions,					/* actions */
    XtNumber(actions),				/* num_actions */
    resources,					/* resources */
    XtNumber(resources),			/* num_resources */
    NULLQUARK,					/* xrm_class */
    TRUE,					/* compress_motion */
    TRUE,					/* compress_exposure */
    TRUE,					/* compress_enterleave */
    FALSE,					/* visible_interest */
    NULL,					/* destroy */
    Resize,					/* resize */
    Redisplay,					/* expose */
    SetValues,					/* set_values */
    NULL,					/* set_values_hook */
    XtInheritSetValuesAlmost,			/* set_values_almost */
    NULL,					/* get_values_hook */
    NULL,					/* accept_focus */
    XtVersion,					/* version */
    NULL,					/* callback_private */
    defaultTranslations,			/* tm_table */
    QueryGeometry,				/* query_geometry */
    XtInheritDisplayAccelerator,		/* display_accelerator */
    NULL					/* extension */
  },
  {
  /* Control class part */
    0,						/* dummy_field */
  },
  {
  /* SimpleButton class part */
    0,						/* dummy_field */
  },
};

WidgetClass xcSimpleButtonWidgetClass = (WidgetClass)&simpleButtonClassRec;


/* Widget method function definitions */

/*******************************************************************
 NAME:		Initialize.		
 DESCRIPTION:
   This is the initialize method for the SimpleButton widget.  It 
validates user-modifiable instance resources and initializes private 
widget variables and structures.  This function also creates any server 
resources (i.e., GCs, fonts, Pixmaps, etc.) used by this widget.  This
method is called by Xt when the application calls XtCreateWidget().

*******************************************************************/

static void Initialize(request, new)
SimpleButtonWidget request, new;
{
/* Local variables */
Display *display = XtDisplay(new);
int scr = DefaultScreen(XtDisplay(new));
int min_width, min_height;


DPRINTF(("SimpleButton: executing Initialize...\n"));

/*
 * Validate public instance variable settings.
 */

/* Initialize the SimpleButton width and height. */
   min_width = XTextWidth(new->control.font, new->control.label, 
		strlen(new->control.label)) + (2 * new->control.shade_depth);
   min_height = new->control.font->ascent + new->control.font->descent +
					(2 * new->control.shade_depth);
   if (new->core.width < min_width)
      new->core.width = min_width;
   if (new->core.height < min_height)
      new->core.height = min_height; 
   
/* Initialize private instance variables.  */
   new->simpleButton.armed = False;


/* Set the initial geometry of the SimpleButton elements. */
   Resize(new);


DPRINTF(("SimpleButton: done Initialize\n"));

}  /* end of Initialize */





/*******************************************************************
 NAME:		Redisplay.	
 DESCRIPTION:
   This function is the SimpleButton's Expose method.  It redraws the
SimpleButton's 3D rectangle background, and label string. All drawing 
takes place within the widget's window (no need for an off-screen pixmap).

*******************************************************************/

static void Redisplay(w, event)
SimpleButtonWidget w;
XExposeEvent *event;
{

/*
 * Check to see whether or not the widget's window is mapped.  You can't
 * draw into a window that is not mapped.  Realizing a widget doesn't 
 * mean its mapped, but this call will work for most Window Managers.
 */
   if (!XtIsRealized(w))
      return;

DPRINTF(("SimpleButton: executing Redisplay\n"));

   Draw_button(w, w->control.gc, RAISED);

DPRINTF(("SimpleButton: done Redisplay\n"));

}  /* end of Redisplay */





/*******************************************************************
 NAME:		SetValues.
 DESCRIPTION:
   This is the set_values method for this widget. It validates resource
settings set with XtSetValues. If a resource is changed that would
require re-drawing the widget, return True.

*******************************************************************/

static Boolean SetValues(cur, req, new)
SimpleButtonWidget cur, req, new;
{
/* Local variables */
Boolean do_redisplay = False;
int min_width, min_height;

DPRINTF(("SimpleButton: executing SetValues for %ld\n", cur));

/* Validate new resource settings. */

/* Check if the label has changed */
   if (strcmp(new->control.label, cur->control.label) != 0)
   {
      min_width = XTextWidth(new->control.font, new->control.label, 
		strlen(new->control.label)) + (2 * new->control.shade_depth);
      if (new->core.width < min_width)
         new->core.width = min_width;

      do_redisplay = True;
   }

/* Check if font has changed, set min width and height if it has. */
   if (new->control.font != cur->control.font)
   {
      min_width = XTextWidth(new->control.font, new->control.label, 
		strlen(new->control.label)) + (2 * new->control.shade_depth);
      min_height = new->control.font->ascent + new->control.font->descent +
					(2 * new->control.shade_depth);
      if (new->core.width < min_width)
         new->core.width = min_width;
      if (new->core.height < min_height)
         new->core.height = min_height; 

      do_redisplay = True;
   }
   
   if (do_redisplay == True)
      Resize(new);

DPRINTF(("SimpleButton: done SetValues\n"));
   return do_redisplay;


}  /* end of SetValues */




/*******************************************************************
 NAME:		Resize.
 DESCRIPTION:
   This is the resize method of the SimpleButton widget. It resizes the
SimpleButton's graphics based on the new width and height of the widget's
window.  

*******************************************************************/

static void Resize(w)
SimpleButtonWidget w;
{

DPRINTF(("SimpleButton: executing Resize\n"));
/* Set the widgets new width and height. */
   w->simpleButton.face.x = w->simpleButton.face.y = w->control.shade_depth;
   w->simpleButton.face.width = w->core.width - (2*w->control.shade_depth)+1;
   w->simpleButton.face.height = w->core.height - (2*w->control.shade_depth)+1;


/* Set the new label location.  Center it in the button. */
   w->simpleButton.lbl.x = (w->core.width / 2 + 0.5)
		- (XTextWidth(w->control.font, w->control.label, 
			strlen(w->control.label)) / 2 + 0.5); 
   w->simpleButton.lbl.y =  (w->core.height / 2 + 0.5) 
		+ ((w->control.font->ascent + 
			w->control.font->descent) / 2)
		- w->control.font->descent; 

DPRINTF(("SimpleButton: done Resize\n"));

}  /* end of Resize */




/*******************************************************************
 NAME:		QueryGeometry.		
 DESCRIPTION:
   This function is the widget's query_geometry method.  It simply
checks the proposed size and returns the appropriate value based on
the proposed size.  If the proposed size is greater than the maximum
appropriate size for this widget, QueryGeometry returns the recommended
size.

*******************************************************************/

static XtGeometryResult QueryGeometry(w, proposed, answer)
SimpleButtonWidget w;
XtWidgetGeometry *proposed, *answer;
{
/* Set the request mode mask for the returned answer. */
   answer->request_mode = CWWidth | CWHeight;

/* Set the recommended size. */
   answer->width = (w->core.width > MAX_SB_WIDTH)
	? MAX_SB_WIDTH : w->core.width;
   answer->height = (w->core.height > MAX_SB_HEIGHT)
	? MAX_SB_HEIGHT : w->core.height;

/* 
 * Check the proposed dimensions. If the proposed size is larger than
 * appropriate, return the recommended size.
 */
   if (((proposed->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight))
	&& proposed->width == answer->width 
	&& proposed->height == answer->height)
      return XtGeometryYes;
   else if (answer->width == w->core.width && answer->height == w->core.height)
      return XtGeometryNo;
   else
      return XtGeometryAlmost;

}  /* end of QueryGeometry */





/* Widget action functions. */

/*******************************************************************
 NAME:		Do_button.		
 DESCRIPTION:
   This function activates the OK Button and calls the application's
registered callback function.

*******************************************************************/

static void Do_button(w, event)
SimpleButtonWidget w;
XEvent *event;
{
/* Local variables */

/*
 * ButtonRelease event. Call the application's callback function
 * and pass the value's data type and new value.
 */
   if ((w->simpleButton.armed) && (event->type == ButtonRelease))
   {
      Draw_button(w, w->control.gc, RAISED);
      w->simpleButton.armed = False;

   /* Call the callback functions. */
      if (w->simpleButton.callback != NULL)
         XtCallCallbacks(w, XcNactivateCallback, NULL);
   }
/*
 * ButtonPress event. Just depress the button and arm it.
 */
   else if (event->type == ButtonPress)
   {
      w->simpleButton.armed = True;
      Draw_button(w, w->control.gc, DEPRESSED);
   }

}  /* end of Do_button */




/*******************************************************************
 NAME:		Do_enter_leave.		
 DESCRIPTION:
   This function handles Enter and Leave Notify events.

*******************************************************************/

static void Do_enter_leave(w, event)
SimpleButtonWidget w;
XEvent *event;
{
/* Local variables */

/*
 * LeaveNotify event. 
 */
   if ((w->simpleButton.armed) && (event->type == LeaveNotify))
   {
      w->simpleButton.armed = False;
      Draw_button(w, w->control.gc, RAISED);
   }
/*
 * EnterNotify event.
 */
   else if ((event->type == EnterNotify) && 
			(event->xcrossing.state & Button1Mask))
   {
      w->simpleButton.armed = True;
      Draw_button(w, w->control.gc, DEPRESSED);
   }

}  /* end of Do_enter_leave */



/*******************************************************************
 NAME:		Draw_button.		
 DESCRIPTION:
   Redraws the button in the given state, RAISED or DEPRESSED.

*******************************************************************/

static void Draw_button(w, gc, type)
SimpleButtonWidget w;
GC gc;
Type3d type;
{

/* Draw the 3D rectangle background for the SimpleButton in the given type. */
   XSetClipMask(XtDisplay(w), gc, None);
   Rect3d(w, XtDisplay(w), XtWindow(w), gc,
		0, 0, w->core.width, w->core.height, type);


/* Draw the Label string. */
   XSetClipRectangles(XtDisplay(w), gc, 0, 0, 
  			&(w->simpleButton.face), 1, Unsorted); 
   XSetForeground(XtDisplay(w), gc,
                        w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), gc,
	w->simpleButton.lbl.x, w->simpleButton.lbl.y, 
	w->control.label, strlen(w->control.label));


}  /* end of Draw_button */



/* end of SimpleButton.c */

