/*******************************************************************
 FILE:		Slider.c
 CONTENTS:	Definitions for structures, methods, and actions of the 
		Slider widget.
 AUTHOR:	Paul D. Johnston
 HISTORY:
 Date		Action
 ---------	------------------------------------
 5/23/92	Changed the widget class name so that it is preceded
		by 'xc' with the first major word capitalized.
 3/26/92	Created.

********************************************************************/

#include <stdio.h>

/* Xlib includes */
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

/* Xt includes */
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>


/* Widget includes */
#include <X11/Xc/Xc.h>
#include <X11/Xc/Control.h>
#include <X11/Xc/Value.h>
#include <X11/Xc/SliderP.h>	/* (includes Slider.h also) */


/* Macro redefinition for offset. */
#define offset(field) XtOffset(SliderWidget, field)


/* Declare widget methods */
static void ClassInitialize();
static void Initialize();
static void Redisplay();
static void Resize();
static XtGeometryResult QueryGeometry();
static Boolean SetValues();

/* Declare widget actions */
static void HandleButtons();	/* Handles all ButtonPress/ButtonRelease 
				 * events in the Slider widget.
				 */
static void HandleMotion();	/* Handles all BtnMotion events in the
				 * Slider widget.
				 */
static void HandleKeyPress();	/* Handles all KeyPress events in the
				 * Slider widget.
				 */


/* Declare functions and variables private to this widget */
static void Update_value();
static void Draw_value();
static void Draw_zoom();
static void Set_nob();
static void Call_callback();
static void Set_zoom_window();
static int Power();




/* Define the widget's resource list */
static XtResource resources[] =
{
  {
    XcNorient,
    XcCOrient,
    XcROrient,
    sizeof(XcOrient),
    offset(slider.orient),
    XtRString,
    "vertical"
  },
  {
    XcNnobColor,
    XtCColor,
    XtRPixel,
    sizeof(Pixel),
    offset(slider.nob_pixel),
    XtRString,
    XtDefaultForeground
  },
  {
    XcNmaxZoomFactor,
    XcCMaxZoomFactor,
    XtRInt,
    sizeof(int),
    offset(slider.max_factor),
    XtRImmediate,
    (XtPointer)0
  },
  {
    XcNenableAccept,
    XcCEnableAccept,
    XtRBoolean,
    sizeof(Boolean),
    offset(slider.accept_enabled),
    XtRString,
    "True"
  },
};


/* Widget translation and actions tables. */
static XtActionsRec actions[] =
{
   { "HandleButtons", HandleButtons },
   { "HandleMotion", HandleMotion },
   { "HandleKeyPress", HandleKeyPress },
};

static char defaultTranslations[] = "<Btn1Down>:  HandleButtons() \n\
				     <BtnMotion>:  HandleMotion() \n\
				     <Key>:  HandleKeyPress() \n\
				     <Btn1Up>:  HandleButtons() ";


/* Widget Class Record initialization */
SliderClassRec sliderClassRec =
{
  {
  /* core_class part */
    (WidgetClass) &valueClassRec,		/* superclass */
    "Slider",					/* class_name */
    sizeof(SliderRec),				/* widget_size */
    ClassInitialize,				/* class_initialize */
    NULL,					/* class_part_initialize */
    FALSE,					/* class_inited */
    Initialize,					/* initialize */
    NULL,					/* initialize_hook */
    XtInheritRealize,				/* realize */
    actions,					/* actions */
    XtNumber(actions),				/* num_actions */
    resources,					/* resources */
    XtNumber(resources),			/* num_resources */
    NULLQUARK,					/* xrm_class */
    TRUE,					/* compress_motion */
    TRUE,					/* compress_exposure */
    TRUE,					/* compress_enterleave */
    TRUE,					/* visible_interest */
    NULL,					/* destroy */
    Resize,					/* resize */
    Redisplay,					/* expose */
    SetValues,					/* set_values */
    NULL,					/* set_values_hook */
    XtInheritSetValuesAlmost,			/* set_values_almost */
    NULL,					/* get_values_hook */
    NULL,					/* accept_focus */
    XtVersion,					/* version */
    NULL,					/* callback_private */
    defaultTranslations,			/* tm_table */
    QueryGeometry,				/* query_geometry */
    NULL,					/* display_accelerator */
    NULL,					/* extension */
  },
  {
  /* Control class part */
    0,						/* dummy_field */
  },
  {
  /* Value class part */
    0,						/* dummy_field */
  },
  {
  /* Slider class part */
    0,						/* dummy_field */
  }
};

WidgetClass xcSliderWidgetClass = (WidgetClass)&sliderClassRec;


/* Macro used to calculate the zoom window bounds around the current value */
#define CALC_ZOOM_BND(w, mem ) \
{ w->slider.upper_zoom = (((w->value.val.mem + (w->slider.zoom_window / 2)) > \
    w->value.upper_bound.mem) ? w->value.upper_bound.mem : \
    w->value.val.mem + (w->slider.zoom_window / 2)); \
  w->slider.lower_zoom = (((w->value.val.mem - (w->slider.zoom_window / 2)) < \
    w->value.lower_bound.mem) ? w->value.lower_bound.mem : \
    w->value.val.mem - (w->slider.zoom_window / 2)); } \

/* Simple macro to determine the Zoom Button orientation to use. */
#define GET_ARROW_ORIENT( w, first, second ) \
 ((w->slider.orient == XcVert) ? first : second) \

/* This macro is used to limit the Nob's movement within the channel. */
#define FILTER_POS( w, coord, dim, pos ) \
 ((pos > (w->slider.chnl.coord + w->slider.chnl.dim - (w->slider.nob.dim/2))) \
   ? (w->slider.chnl.coord + w->slider.chnl.dim - (w->slider.nob.dim/2)) : \
  ((pos < (w->slider.chnl.coord + (w->slider.nob.dim/2))) ? \
      (w->slider.chnl.coord + (w->slider.nob.dim/2)) : pos)) \


/* Widget method function definitions */

/*******************************************************************
 NAME:		ClassInitialize.	
 DESCRIPTION:
   This method initializes the Slider widget class. Specifically,
it registers resource value converter functions with Xt.

*******************************************************************/

static void ClassInitialize()
{

   XtAddConverter(XtRString, XcROrient, CvtStringToOrient, NULL, 0);

}  /* end of ClassInitialize */






/*******************************************************************
 NAME:		Initialize.		
 DESCRIPTION:
   This is the initialize method for the Slider widget.  It 
validates user-modifiable instance resources and initializes private 
widget variables and structures.  This function also creates any server 
resources (i.e., GCs, fonts, Pixmaps, etc.) used by this widget.  This
method is called by Xt when the application calls XtCreateWidget().

*******************************************************************/

static void Initialize(request, new)
SliderWidget request, new;
{
/* Local variables */
Display *display = XtDisplay(new);


DPRINTF(("Slider: executing Initialize...\n"));

/*
 * Validate public instance variable settings.
 */
/* Check orientation resource setting. */
   if ((new->slider.orient != XcVert) && (new->slider.orient != XcHoriz))
   {
      XtWarning("Slider: invalid orientation setting");
      new->slider.orient = XcVert;
   }

/* Check maxZoomFactor resource setting. */
   if (new->slider.max_factor > MAX_ZOOM_FACTOR)
   {
      XtWarning("Slider: invalid maxZoomFactor setting");
      new->slider.max_factor = MAX_ZOOM_FACTOR;
   }


/* Initialize the Slider width and height. */
   if (new->core.width < MIN_SL_WIDTH)
      new->core.width = MIN_SL_WIDTH; 
   if (new->core.height < MIN_SL_HEIGHT)
      new->core.height = MIN_SL_HEIGHT; 
   
/* Initialize private instance variables.  */
   new->slider.nob_selected = False;
   new->slider.zoom_factor = 0;

   if ((new->value.datatype == XcLval) || (new->value.datatype == XcHval))
   {
      new->slider.zoom_window = (float)((new->value.upper_bound.lval - 
		new->value.lower_bound.lval) / 
				Power(2, new->slider.zoom_factor));
      new->slider.upper_zoom = new->value.upper_bound.lval;
      new->slider.lower_zoom = new->value.lower_bound.lval;
   }
   else if (new->value.datatype == XcFval)
   {
      new->slider.zoom_window = ((new->value.upper_bound.fval - 
		new->value.lower_bound.fval) / 
				Power(2, new->slider.zoom_factor));
      new->slider.upper_zoom = new->value.upper_bound.fval;
      new->slider.lower_zoom = new->value.lower_bound.fval;
   }

   if (new->value.datatype == XcLval)
      new->slider.adjustment.lval = 1L;
   else if (new->value.datatype == XcHval)
      new->slider.adjustment.lval = 0x1L;
   else if (new->value.datatype == XcFval)
   {
      if (new->value.decimals == 1)
         new->slider.adjustment.fval = 0.1;
      else if (new->value.decimals == 2)
         new->slider.adjustment.fval = 0.01;
      else if (new->value.decimals == 3)
         new->slider.adjustment.fval = 0.001;
      else if (new->value.decimals == 4)
         new->slider.adjustment.fval = 0.0001;
      else if (new->value.decimals == 5)
         new->slider.adjustment.fval = 0.00001;
      else if (new->value.decimals == 6)
         new->slider.adjustment.fval = 0.000001;
   }


/* Set the initial geometry of the Slider elements. */
   Resize(new);

/* Establish the initial position of the Slider Nob */
   Set_nob(new);


DPRINTF(("Slider: done Initialize\n"));

}  /* end of Initialize */





/*******************************************************************
 NAME:		Redisplay.	
 DESCRIPTION:
   This function is the Slider's Expose method.  It redraws the
Slider's 3D rectangle background, Value Box, label and value string,
and the Slider Nob and Channel.  All drawing takes place within the 
widget's window (no need for an off-screen pixmap).

*******************************************************************/

static void Redisplay(w, event)
SliderWidget w;
XExposeEvent *event;
{
/* Local variables */
char zoom[3];
ArrowType orientation;

/*
 * Check to see whether or not the widget's window is mapped.  You can't
 * draw into a window that is not mapped.  Realizing a widget doesn't 
 * mean its mapped, but this call will work for most Window Managers.
 */
   if (!XtIsRealized(w))
      return;

DPRINTF(("Slider: executing Redisplay\n"));

/* Draw the 3D rectangle background for the Slider. */
   XSetClipMask(XtDisplay(w), w->control.gc, None);
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		0, 0, w->core.width, w->core.height, RAISED);

/* Draw the Label string. */
   XSetClipRectangles(XtDisplay(w), w->control.gc, 0, 0, 
  			&(w->slider.face), 1, Unsorted); 
   XSetForeground(XtDisplay(w), w->control.gc,
                        w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), w->control.gc,
	w->slider.lbl.x, w->slider.lbl.y, 
	w->control.label, strlen(w->control.label));

/* Draw the Zoom Buttons */
   if (w->slider.max_factor > 0)
   {
      orientation = GET_ARROW_ORIENT(w, UP, RIGHT);
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.up_arrow, orientation, RAISED);
      orientation = GET_ARROW_ORIENT(w, DOWN, LEFT);
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.down_arrow, orientation, RAISED);
      Draw_zoom(w);
   }

/* Draw the Accept Button if it is enabled. */
   if (w->slider.accept_enabled == True)
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.acpt.x, w->slider.acpt.y, 
		w->slider.acpt.width, w->slider.acpt.height, RAISED);

/* Draw the Slider Channel */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->slider.chnl.x - w->control.shade_depth, 
	w->slider.chnl.y - w->control.shade_depth,
	w->slider.chnl.width + (2 * w->control.shade_depth),  
	w->slider.chnl.height + (2 * w->control.shade_depth), DEPRESSED);

/* Draw the Slider Nob */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->slider.nob.x, w->slider.nob.y,
	w->slider.nob.width, w->slider.nob.height, RAISED); 
   XSetForeground(XtDisplay(w), w->control.gc, w->slider.nob_pixel);
   if ((w->slider.nob.width > (2 * w->control.shade_depth)) && 
			(w->slider.nob.height > (2 * w->control.shade_depth)))
   {
      XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x + w->control.shade_depth, 
		w->slider.nob.y + w->control.shade_depth,
		w->slider.nob.width - (2 * w->control.shade_depth),  
		w->slider.nob.height - (2 * w->control.shade_depth));
   }

/* Draw the Value Box */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->value.value_box.x - w->control.shade_depth, 
	w->value.value_box.y - w->control.shade_depth,
	w->value.value_box.width + (2 * w->control.shade_depth),  
	w->value.value_box.height + (2 * w->control.shade_depth), 
							DEPRESSED);

/* Draw the new value string in the Value Box. */
   Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);


DPRINTF(("Slider: done Redisplay\n"));

}  /* end of Redisplay */





/*******************************************************************
 NAME:		SetValues.
 DESCRIPTION:
   This is the set_values method for this widget. It validates resource
settings set with XtSetValues. If a resource is changed that would
require re-drawing the widget, return True.

*******************************************************************/

static Boolean SetValues(cur, req, new)
SliderWidget cur, req, new;
{
/* Local variables */
Boolean do_redisplay = False;


DPRINTF(("Slider: executing SetValues \n"));

/* Validate new resource settings. */

/* Check orientation resource setting. */
   if (new->slider.orient != cur->slider.orient)
   {
      do_redisplay = True;
      if ((new->slider.orient != XcVert) && (new->slider.orient != XcHoriz))
      {
         XtWarning("Slider: invalid orientation setting");
         new->slider.orient = XcVert;
      }
   }

/* Check maxZoomFactor resource setting. */
   if (new->slider.max_factor != cur->slider.max_factor)
   {
      do_redisplay = True;
      if (new->slider.max_factor > MAX_ZOOM_FACTOR)
      {
         XtWarning("Slider: invalid maxZoomFactor setting");
         new->slider.max_factor = MAX_ZOOM_FACTOR;
      }
   }

/* Check enableAccept resource setting. */
   if (new->slider.accept_enabled != cur->slider.accept_enabled)
   {
      do_redisplay = True;
      if ((new->slider.accept_enabled != True) && 
			(new->slider.accept_enabled != False))
      {
         XtWarning("Slider: invalid enableAccept setting");
         new->slider.accept_enabled = True;
      }
   }

/*
 * Update the fine adjustment value based on the datatype and decimals
 * setting.
 */
   if (new->value.datatype != cur->value.datatype)
   {
      if (new->value.datatype == XcLval)
         new->slider.adjustment.lval = 1L;
      else if (new->value.datatype == XcHval)
         new->slider.adjustment.lval = 0x1L;
      else if (new->value.datatype == XcFval)
      {
         if (new->value.decimals == 1)
            new->slider.adjustment.fval = 0.1;
         else if (new->value.decimals == 2)
            new->slider.adjustment.fval = 0.01;
         else if (new->value.decimals == 3)
            new->slider.adjustment.fval = 0.001;
         else if (new->value.decimals == 4)
            new->slider.adjustment.fval = 0.0001;
         else if (new->value.decimals == 5)
            new->slider.adjustment.fval = 0.00001;
         else if (new->value.decimals == 6)
            new->slider.adjustment.fval = 0.000001;
      }
   }

   if ((((new->value.datatype == XcLval) || (new->value.datatype == XcHval)) && 
		(new->value.val.lval != cur->value.val.lval))
      || ((new->value.datatype == XcFval) && 
		(new->value.val.fval != cur->value.val.fval)))
   {
      do_redisplay = True;
      Set_nob(new);
   }


DPRINTF(("Slider: done SetValues\n"));
   return do_redisplay;


}  /* end of SetValues */




/*******************************************************************
 NAME:		Resize.
 DESCRIPTION:
   This is the resize method of the Slider widget. It resizes the
Slider's graphics based on the new width and height of the widget's
window.  

*******************************************************************/

static void Resize(w)
SliderWidget w;
{

DPRINTF(("Slider: executing Resize\n"));

/* Set the widgets new width and height. */
   w->slider.face.x = w->slider.face.y = w->control.shade_depth;
   w->slider.face.width = w->core.width - (2*w->control.shade_depth)+1;
   w->slider.face.height = w->core.height - (2*w->control.shade_depth)+1;

/* Establish the new Value Box geometry. */
   if (w->slider.orient == XcVert)
   {
      w->value.value_box.x = w->slider.face.x + (2*w->control.shade_depth);
      w->value.value_box.y = w->slider.face.y + w->slider.face.height -
			(short)(w->slider.face.height / 8 + 0.5) +
					(2 * w->control.shade_depth);
      w->value.value_box.width = w->slider.face.width -
					(4 * w->control.shade_depth);
      w->value.value_box.height = (short)(w->slider.face.height / 8 +0.5) -
					(2*w->control.shade_depth);
   }
   else
   {
      w->value.value_box.x = w->slider.face.x + 
				(short)(w->slider.face.width / 3 + 0.5) +
				w->control.shade_depth;
      w->value.value_box.y = w->slider.face.y +
			(short)((2 * w->slider.face.height) / 3 + 0.5) +
					(2 * w->control.shade_depth);
      w->value.value_box.width = (short)(w->slider.face.width / 3 +0.5) -
					(2 * w->control.shade_depth);
      w->value.value_box.height = (short)(w->slider.face.height / 3 +0.5) -
					(2*w->control.shade_depth);
   }

/* Set the position of the value string within the Value Box. */
   Position_val(w);

/* Set the new label location. */
   w->slider.lbl.x = (w->core.width / 2 + 0.5) -
		(XTextWidth(w->control.font, w->control.label, 
			strlen(w->control.label)) / 2 + 0.5); 
   w->slider.lbl.y = w->slider.face.y + w->control.font->ascent + 
			(short)(w->slider.face.height / 32 + 0.5);


/* Resize the Channel */
   if (w->slider.orient == XcVert)
   {
      w->slider.chnl.x = w->slider.face.x +
			(short)(w->slider.face.width / 3 +0.5) +
				w->control.shade_depth;
      w->slider.chnl.y = w->slider.face.y +
			(short)(w->slider.face.height / 8 +0.5) +
				w->control.shade_depth;
      w->slider.chnl.width = (short)(w->slider.face.width / 3 +0.5) -
					(2*w->control.shade_depth);
      w->slider.chnl.height = (short)((6*w->slider.face.height) / 8 +0.5) -
					(2*w->control.shade_depth);
   }
   else
   {
      w->slider.chnl.x = w->slider.face.x +
			(short)(w->slider.face.width / 8 +0.5) +
				w->control.shade_depth;
      w->slider.chnl.y = w->slider.face.y +
			(short)(w->slider.face.height / 3 +0.5) +
				w->control.shade_depth;
      w->slider.chnl.width = (short)((6 * w->slider.face.width) / 8 +0.5) -
					(2*w->control.shade_depth);
      w->slider.chnl.height = (short)(w->slider.face.height / 3 +0.5) -
					(2*w->control.shade_depth);
   }


/* Resize the Zoom Buttons and Zoom Factor rectangle. */
   if (w->slider.max_factor > 0)
   {
      if (w->slider.orient == XcVert)
      {
         w->slider.up_arrow.x = w->slider.chnl.x + w->slider.chnl.width +
					(2 * w->control.shade_depth);
         w->slider.up_arrow.y = w->slider.face.y + 
				(w->slider.face.height / 3 + 0.5);
         w->slider.up_arrow.width = (w->slider.face.width / 3 + 0.5) -
					w->control.shade_depth;
         w->slider.up_arrow.height = (w->slider.face.height / 9 + 0.5);

         w->slider.zoom.x = w->slider.up_arrow.x;
         w->slider.zoom.y = w->slider.up_arrow.y + w->slider.up_arrow.height;
         w->slider.zoom.width = w->slider.up_arrow.width;
         w->slider.zoom.height = w->slider.up_arrow.height;

         w->slider.down_arrow.x = w->slider.up_arrow.x;
         w->slider.down_arrow.y = w->slider.zoom.y + w->slider.zoom.height;
         w->slider.down_arrow.width = w->slider.up_arrow.width;
         w->slider.down_arrow.height = w->slider.up_arrow.height;
      }
      else
      {
         w->slider.down_arrow.x = w->value.value_box.x + 
				w->value.value_box.width +
					(2 * w->control.shade_depth);
         w->slider.down_arrow.y = w->value.value_box.y - 
						w->control.shade_depth; 
         w->slider.down_arrow.width = (w->slider.face.width / 9 + 0.5); 
         w->slider.down_arrow.height = w->value.value_box.height;

         w->slider.zoom.x = w->slider.down_arrow.x + w->slider.down_arrow.width;
         w->slider.zoom.y = w->slider.down_arrow.y;
         w->slider.zoom.width = w->slider.down_arrow.width;
         w->slider.zoom.height = w->slider.down_arrow.height;

         w->slider.up_arrow.x = w->slider.zoom.x + w->slider.zoom.width;
         w->slider.up_arrow.y = w->slider.down_arrow.y;
         w->slider.up_arrow.width = w->slider.down_arrow.width;
         w->slider.up_arrow.height = w->slider.down_arrow.height;
      }
   }

/* Resize the Accept Button */
   if (w->slider.accept_enabled == True)
   {
      if (w->slider.orient == XcVert)
      {
         w->slider.acpt.x = w->slider.face.x; 
         w->slider.acpt.y = w->slider.face.y + 
				(4 * (w->slider.face.height / 9) + 0.5);
         w->slider.acpt.width = (w->slider.face.width / 3 + 0.5) -
					w->control.shade_depth;
         w->slider.acpt.height = (w->slider.face.height / 9 + 0.5);
      }
      else if (w->slider.orient == XcHoriz)
      {
         w->slider.acpt.x = w->slider.chnl.x; 
         w->slider.acpt.y = w->value.value_box.y - w->control.shade_depth; 
         w->slider.acpt.width = (w->slider.face.width / 9 + 0.5);
         w->slider.acpt.height = (w->slider.face.height / 3 + 0.5) -
						w->control.shade_depth;
      }
   }


/* Resize the Slider Nob. */
   if (w->slider.orient == XcVert)
   {
      w->slider.nob.width = w->slider.chnl.width;
      w->slider.nob.height = (short)(w->slider.chnl.height / 10 + 0.5);
   }
   else
   {
      w->slider.nob.width = (short)(w->slider.chnl.width / 10 + 0.5); 
      w->slider.nob.height = w->slider.chnl.height;
   }

   Set_nob(w);

DPRINTF(("Slider: done Resize\n"));

}  /* end of Resize */




/*******************************************************************
 NAME:		QueryGeometry.		
 DESCRIPTION:
   This function is the widget's query_geometry method.  It simply
checks the proposed size and returns the appropriate value based on
the proposed size.  If the proposed size is greater than the maximum
appropriate size for this widget, QueryGeometry returns the recommended
size.

*******************************************************************/

static XtGeometryResult QueryGeometry(w, proposed, answer)
SliderWidget w;
XtWidgetGeometry *proposed, *answer;
{
/* Set the request mode mask for the returned answer. */
   answer->request_mode = CWWidth | CWHeight;

/* Set the recommended size. */
   answer->width = (w->core.width > MAX_SL_WIDTH)
	? MAX_SL_WIDTH : w->core.width;
   answer->height = (w->core.height > MAX_SL_HEIGHT)
	? MAX_SL_HEIGHT : w->core.height;

/* 
 * Check the proposed dimensions. If the proposed size is larger than
 * appropriate, return the recommended size.
 */
   if (((proposed->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight))
	&& proposed->width == answer->width 
	&& proposed->height == answer->height)
      return XtGeometryYes;
   else if (answer->width == w->core.width && answer->height == w->core.height)
      return XtGeometryNo;
   else
      return XtGeometryAlmost;

}  /* end of QueryGeometry */





/* Widget action functions. */

/*******************************************************************
 NAME:		HandleButtons.
 DESCRIPTION:
   This function handles ButtonPress and ButtonRelease events in the
Slider widget.

*******************************************************************/

static void HandleButtons(w, event)
SliderWidget w;
XEvent *event;
{
/* Local variables */
ArrowType orientation;
int new_pos;


   if (event->type == ButtonRelease) 
   {
      if (w->slider.nob_selected) 
      {
         w->slider.nob_selected = False;
	 if (w->slider.accept_enabled == False)
            Call_callback(w);
      }
      else if (w->slider.acpt_armed)
      {
	 w->slider.acpt_armed = False;
         Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.acpt.x, w->slider.acpt.y, 
		w->slider.acpt.width, w->slider.acpt.height, RAISED);
	 if (Point_In_Rect(event->xbutton.x, event->xbutton.y, 
		w->slider.acpt.x, w->slider.acpt.y, 
		w->slider.acpt.width, w->slider.acpt.height)) 
            Call_callback(w);
      }
      else if (w->slider.up_armed) 
      {
	 w->slider.up_armed = False;
	 orientation = GET_ARROW_ORIENT(w, UP, RIGHT);
	 Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.up_arrow, orientation, RAISED);

	 if (Point_In_Rect(event->xbutton.x, event->xbutton.y, 
		w->slider.up_arrow.x, w->slider.up_arrow.y, 
		w->slider.up_arrow.width, w->slider.up_arrow.height)) 
	 {
	 /* Set the zoom factor and zoom window */
	    w->slider.zoom_factor++;
	    if (w->slider.zoom_factor > w->slider.max_factor)
	       w->slider.zoom_factor = w->slider.max_factor;

	    Draw_zoom(w);
	    Set_zoom_window(w);

         /* Erase the current Nob */
            XSetForeground(XtDisplay(w), w->control.gc, 
					w->control.background_pixel);
            XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x, w->slider.nob.y,
		w->slider.nob.width+1, w->slider.nob.height+1);

	    Set_nob(w);
         }
      }
      else if (w->slider.down_armed) 
      {
	 w->slider.down_armed = False;
	 orientation = GET_ARROW_ORIENT(w, DOWN, LEFT);
	 Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.down_arrow, orientation, RAISED);

	 if (Point_In_Rect(event->xbutton.x, event->xbutton.y, 
		w->slider.down_arrow.x, w->slider.down_arrow.y, 
		w->slider.down_arrow.width, w->slider.down_arrow.height)) 
	 {
	 /* Set the zoom factor and zoom window */
	    w->slider.zoom_factor--;
	    if (w->slider.zoom_factor < 0)
	       w->slider.zoom_factor = 0;

	    Draw_zoom(w);
	    Set_zoom_window(w);

         /* Erase the current Nob */
            XSetForeground(XtDisplay(w), w->control.gc, 
					w->control.background_pixel);
            XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x, w->slider.nob.y,
		w->slider.nob.width+1, w->slider.nob.height+1);

	    Set_nob(w);
         }
      }
   }

/* ButtonPress event */
   else if (event->type == ButtonPress) 
   {
      if (Point_In_Rect(event->xbutton.x, event->xbutton.y, w->slider.nob.x,
		w->slider.nob.y, w->slider.nob.width, w->slider.nob.height)) 
      {
         w->slider.nob_selected = True;
      }
      else if (Point_In_Rect(event->xbutton.x, event->xbutton.y, 
		w->slider.chnl.x, w->slider.chnl.y, 
		w->slider.chnl.width, w->slider.chnl.height))
      {
      /* Update the widget's new value based on the pointer position */
         if (w->slider.orient == XcVert)
	    new_pos = FILTER_POS(w, y, height, event->xbutton.y);
         else if (w->slider.orient == XcHoriz)
	    new_pos = FILTER_POS(w, x, width, event->xbutton.x);
         Update_value(w, new_pos); 

      /* Erase the current Nob */
         XSetForeground(XtDisplay(w), w->control.gc, w->control.background_pixel);
         XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x, w->slider.nob.y,
		w->slider.nob.width+1, w->slider.nob.height+1);

      /* Set the Nob's new position */
         Set_nob(w);

         Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);

	 if (w->slider.accept_enabled == False)
            Call_callback(w);
      }
      else if ((w->slider.max_factor > 0) && (Point_In_Rect(event->xbutton.x, 
		event->xbutton.y, w->slider.up_arrow.x, w->slider.up_arrow.y, 
		w->slider.up_arrow.width, w->slider.up_arrow.height)))
      {
	 w->slider.up_armed = True;
	 orientation = GET_ARROW_ORIENT(w, UP, RIGHT);
	 Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.up_arrow, orientation, DEPRESSED);
      }
      else if ((w->slider.max_factor > 0) && (Point_In_Rect(event->xbutton.x, 
		event->xbutton.y, w->slider.down_arrow.x, 
		w->slider.down_arrow.y, w->slider.down_arrow.width, 
					w->slider.down_arrow.height)))
      {
	 w->slider.down_armed = True;
	 orientation = GET_ARROW_ORIENT(w, DOWN, LEFT);
	 Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
			&w->slider.down_arrow, orientation, DEPRESSED);
      }
      else if ((w->slider.accept_enabled == True) && 
		(Point_In_Rect(event->xbutton.x, event->xbutton.y, 
			w->slider.acpt.x, w->slider.acpt.y, 
			w->slider.acpt.width, w->slider.acpt.height)))
      {
	 w->slider.acpt_armed = True;
         Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.acpt.x, w->slider.acpt.y, 
		w->slider.acpt.width, w->slider.acpt.height, DEPRESSED);
      }
   }

}  /* end of HandleButtons */






/*******************************************************************
 NAME:		HandleMotion.		
 DESCRIPTION:
   This function handles PointerMotion events by updating the value
and setting the Nob position based on the new pointer position.

*******************************************************************/

static void HandleMotion(w, event)
SliderWidget w;
XEvent *event;
{
/* Local variables */
int new_pos;

/* Make sure that the Nob is selected. */
   if ((w->slider.nob_selected) && 
		(Point_In_Rect(event->xmotion.x, event->xmotion.y, 
			w->slider.chnl.x, w->slider.chnl.y, 
			w->slider.chnl.width, w->slider.chnl.height)))
   {
   /* Update the widget's new value based on the pointer position */
      if (w->slider.orient == XcVert)
	 new_pos = FILTER_POS(w, y, height, event->xmotion.y);
      else if (w->slider.orient == XcHoriz)
	 new_pos = FILTER_POS(w, x, width, event->xmotion.x);
      Update_value(w, new_pos); 

   /* Erase the current Nob */
      XSetForeground(XtDisplay(w), w->control.gc, w->control.background_pixel);
      XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x, w->slider.nob.y,
		w->slider.nob.width+1, w->slider.nob.height+1);

   /* Set the Nob's new position */
      Set_nob(w);

      Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);

   }

}  /* end of HandleMotion */



/*******************************************************************
 NAME:		HandleKeyPress.		
 DESCRIPTION:
   This function updates the Slider's Nob and value when the user
presses the up or down arrow keys.

*******************************************************************/

static void HandleKeyPress(w, event)
SliderWidget w;
XKeyEvent *event;
{
/* Local variables */
KeySym key;
Boolean value_changed = False;

/* Only process Up and Down arrow key presses */
   key = XLookupKeysym(event, 0);
   if (key == XK_Up)
   {
      value_changed = True;
   /* Increment the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))   
      {
	 w->value.val.lval += w->slider.adjustment.lval;
	 if (w->value.val.lval > w->value.upper_bound.lval) 
	 {
	    w->value.val.lval = w->value.upper_bound.lval;
	    value_changed = False;
	 }
      }
      else if (w->value.datatype == XcFval)
      {
	 w->value.val.fval += w->slider.adjustment.fval;
	 if (w->value.val.fval > w->value.upper_bound.fval) 
	 {
	    w->value.val.fval = w->value.upper_bound.fval; 
	    value_changed = False;
	 }
      }
   }
   else if (key == XK_Down)
   {
      value_changed = True;
   /* Decrement the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))   
      {
	 w->value.val.lval -= w->slider.adjustment.lval;
	 if (w->value.val.lval < w->value.lower_bound.lval)
	 {
	    w->value.val.lval = w->value.lower_bound.lval;
      	    value_changed = False;
	 }
      }
      else if (w->value.datatype == XcFval)
      {
	 w->value.val.fval -= w->slider.adjustment.fval;
	 if (w->value.val.fval < w->value.lower_bound.fval)
	 {
	    w->value.val.fval = w->value.lower_bound.fval;
	    value_changed = False;
	 }
      }
   }


/* Position the Nob accordingly */
   if (value_changed)
   {
   /* Erase the current Nob */
      XSetForeground(XtDisplay(w), w->control.gc, w->control.background_pixel);
      XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.nob.x, w->slider.nob.y,
		w->slider.nob.width+1, w->slider.nob.height+1);

   /* Set the Nob's new position */
      Set_nob(w);

      Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);

      if (w->slider.accept_enabled == False)
         Call_callback(w);
   }

}  /* end of HandleKeyPress */




/*******************************************************************
 NAME:		Call_callback.		
 DESCRIPTION:
   This function prepares the call_data structure and calls the
application's callback function with this data.

*******************************************************************/

static void Call_callback(w)
SliderWidget w;
{
/* Local variables */
static XcCallData call_data;

/* Prepare the call_data */
   call_data.dtype = w->value.datatype;
   call_data.decimals = w->value.decimals;
   if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))  
      call_data.value.lval = w->value.val.lval;
   else if (w->value.datatype == XcFval)
      call_data.value.fval = w->value.val.fval;

/* Call the application's callback function(s). */
   if (w->value.callback != NULL)
      XtCallCallbacks(w, XcNvalueCallback, &call_data);

}  /* end of Call_callback */




/*******************************************************************
 NAME:		Update_value.		
 DESCRIPTION:
   This function updates the Slider's value variable based on the new
Slider Nob position.

*******************************************************************/

static void Update_value(w, new_pos)
SliderWidget w;
short new_pos;
{
/* Local variables */
float from, dim;
XcVType temp;

   if (w->slider.orient == XcVert)
   {
      from = (float)(w->slider.chnl.y + w->slider.chnl.height - 
				(w->slider.nob.height/2) - new_pos);
      dim = (float)w->slider.chnl.height - w->slider.nob.height;
   }
   else
   {
      from = (float)(new_pos - (w->slider.chnl.x + (w->slider.nob.width/2)));
      dim = (float)w->slider.chnl.width - w->slider.nob.width;
   }


   if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))  
   {
      temp.lval = (long)Correlate(from, dim, 
			(w->slider.upper_zoom - w->slider.lower_zoom));
      w->value.val.lval = (long)(w->slider.lower_zoom + temp.lval >
		w->slider.upper_zoom ? w->slider.upper_zoom :
				w->slider.lower_zoom + temp.lval); 
   }
   else if (w->value.datatype == XcFval)
   {
      temp.fval = Correlate(from, dim, 
			(w->slider.upper_zoom - w->slider.lower_zoom));
      w->value.val.fval = (w->slider.lower_zoom + temp.fval >
		w->slider.upper_zoom ? w->slider.upper_zoom :
				w->slider.lower_zoom + temp.fval); 
   }

}  /* end of Update_value */




/*******************************************************************
 NAME:		Draw_value.		
 DESCRIPTION:
   This function redraws the value string in the Value Box.

*******************************************************************/

static void Draw_value(w, display, drawable, gc)
SliderWidget w;
Display *display;
Drawable drawable;
GC gc;
{
/* Local variables */
char *temp;
XRectangle val_clip;

/* Clear the Value Box by re-filling it with its background color. */
   XSetForeground(display, gc, w->value.value_bg_pixel); 
   XFillRectangle(display, drawable, gc,
  	w->value.value_box.x, w->value.value_box.y, 
  	w->value.value_box.width, w->value.value_box.height); 

/*
 * Now draw the value string in its foreground color, clipped by the
 * Value Box.
 */
   XSetForeground(display, gc, w->value.value_fg_pixel); 
   XSetClipRectangles(display, gc, 0, 0, 
  	&(w->value.value_box), 1, Unsorted); 

   temp = Print_value(w->value.datatype, &w->value.val, w->value.decimals);

   Position_val(w);

   XDrawString(display, drawable, gc,
    	w->value.vp.x, w->value.vp.y, temp, strlen(temp)); 

/* Reset the clip_mask to no clipping. */
   XSetClipMask(display, gc, None);

}  /* end of Draw_value */




/*******************************************************************
 NAME:		Set_nob.		
 DESCRIPTION:
   This function sets the X,Y coordinates of the Slider Nob based
on the new value resource setting.

*******************************************************************/

static void Set_nob(w)
SliderWidget w;
{
/* Local variables */
float dim;
unsigned long temp;

   if (w->slider.orient == XcVert)
      dim = (float)w->slider.chnl.height - w->slider.nob.height;
   else
      dim = (float)w->slider.chnl.width - w->slider.nob.width;

   if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
      temp = Correlate(((float)w->value.val.lval - w->slider.lower_zoom), 
		(w->slider.upper_zoom - w->slider.lower_zoom), dim);
   else if (w->value.datatype == XcFval)
      temp = Correlate((w->value.val.fval - w->slider.lower_zoom),
		(w->slider.upper_zoom - w->slider.lower_zoom), dim);

   if (w->slider.orient == XcVert)
   {
      w->slider.nob.x = w->slider.chnl.x;
      w->slider.nob.y = w->slider.chnl.y + w->slider.chnl.height - temp -
						w->slider.nob.height; 
   }
   else
   {
      w->slider.nob.x = w->slider.chnl.x + temp; 
      w->slider.nob.y = w->slider.chnl.y;
   }

/* Now draw the Nob at its new position. */
   if (XtIsRealized(w))
   {
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	   w->slider.nob.x, w->slider.nob.y,
	   w->slider.nob.width, w->slider.nob.height, RAISED); 
      if ((w->slider.nob.width > (2 * w->control.shade_depth)) && 
			(w->slider.nob.height > (2 * w->control.shade_depth)))
      {
         XSetForeground(XtDisplay(w), w->control.gc, w->slider.nob_pixel);
         XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
	   	w->slider.nob.x + w->control.shade_depth, 
	   	w->slider.nob.y + w->control.shade_depth,
	   	w->slider.nob.width - (2 * w->control.shade_depth),  
	   	w->slider.nob.height - (2 * w->control.shade_depth));
      }
   }

}  /* end of Set_nob */




/*******************************************************************
 NAME:		Draw_zoom.		
 DESCRIPTION:
   This function draws the Zoom factor string between the Zoom buttons.

*******************************************************************/

static void Draw_zoom(w)
SliderWidget w;
{
/* Local variables */
char zoom[10];

/* Re-position the Zoom Factor within the rectangle. */
   sprintf(zoom, "%d", w->slider.zoom_factor);
   w->slider.zoom_lbl.x = w->slider.zoom.x + (w->slider.zoom.width / 2 +0.5) -
		(XTextWidth(w->control.font, zoom, strlen(zoom)) / 2 + 0.5);
   w->slider.zoom_lbl.y = w->slider.zoom.y + (w->slider.zoom.height / 2 +0.5) +
		((w->control.font->ascent + w->control.font->descent) / 2)
					- w->control.font->descent;

/* Draw the Zoom Factor. */
   XSetClipRectangles(XtDisplay(w), w->control.gc, 0, 0, 
  			&(w->slider.zoom), 1, Unsorted); 
   XSetForeground(XtDisplay(w), w->control.gc, w->control.background_pixel);
   XFillRectangle(XtDisplay(w), XtWindow(w), w->control.gc,
		w->slider.zoom.x, w->slider.zoom.y, 
		w->slider.zoom.width+1, w->slider.zoom.height+1); 
   XSetForeground(XtDisplay(w), w->control.gc, w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), w->control.gc,
	w->slider.zoom_lbl.x, w->slider.zoom_lbl.y, zoom, strlen(zoom));

/* Reset the clip_mask to no clipping. */
   XSetClipMask(XtDisplay(w), w->control.gc, None);

}  /* end of Draw_zoom */




/*******************************************************************
 NAME:		Set_zoom_window.		
 DESCRIPTION:
   Establishes the zoom window based on the zoom factor.

*******************************************************************/

static void Set_zoom_window(w)
SliderWidget w;
{

   if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
   {
      w->slider.zoom_window = (float)((w->value.upper_bound.lval - 
		w->value.lower_bound.lval) / Power(2, w->slider.zoom_factor));
      if (w->slider.zoom_factor == 0)
      {
	 w->slider.upper_zoom = w->value.upper_bound.lval;
	 w->slider.lower_zoom = w->value.lower_bound.lval;
      }
      else
	 CALC_ZOOM_BND(w, lval);
   }
   else if (w->value.datatype == XcFval)
   {
      w->slider.zoom_window = ((w->value.upper_bound.fval - 
		w->value.lower_bound.fval) / Power(2, w->slider.zoom_factor));
      if (w->slider.zoom_factor == 0)
      {
	 w->slider.upper_zoom = w->value.upper_bound.fval;
	 w->slider.lower_zoom = w->value.lower_bound.fval;
      }
      else
	 CALC_ZOOM_BND(w, fval);
   }

}  /* end of Set_zoom_window */



/*******************************************************************
 NAME:		Power.		
 DESCRIPTION:
   Returns the value of the specified integer raised to the specified
power.

*******************************************************************/

static int Power(base, exp)
int base, exp;
{
int i, p = 1;

   for (i = 1; i <= exp; ++i)
      p = p * base;

   return p;

}  /* end of Power */


/* end of Slider.c */

