/*******************************************************************
 FILE:		ToggleB.c
 CONTENTS:	Definitions for structures, methods, and actions of the 
		ToggleButton widget.
 AUTHOR:	Paul D. Johnston
 HISTORY:
 Date		Action
 ---------	------------------------------------
 5/23/92	Changed the widget class name so that it is preceded
		by 'xc' with the first major word capitalized.
 10/22/91	Created.

********************************************************************/

#include <stdio.h>

/* Xlib includes */
#include <X11/Xlib.h>

/* Xt includes */
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#ifdef X11R3
#include <X11/Xmu.h>
#else
#include <X11/Xmu/Xmu.h>
#endif


/* Widget includes */
#include <X11/Xc/ToggleBP.h>	/* (includes ToggleB.h also) */


/* Macro redefinition for offset. */
#define offset(field) XtOffset(ToggleButtonWidget, field)


/* Declare widget methods */
static void ClassInitialize();
static void Initialize();
static void Redisplay();
static void Resize();
static XtGeometryResult QueryGeometry();
static Boolean SetValues();

/* Declare widget actions */
static void Do_button();	/* Handles all ButtonDown and ButtonUp 
				 * events in the ToggleButton widget.
				 */
static void Do_enter_leave();	/* Handles all EnterNotify and LeaveNotify
				 * events in the ToggleButton widget.
				 */



/* Declare functions and variables private to this widget */
void CvtStringToToggleState();
void Draw_button();
void Draw_ind_color();
void upstring();


/* Define the widget's resource list */
static XtResource resources[] =
{
  {
    XcNoffColor,
    XcCColor,
    XtRString,
    sizeof(String),
    offset(toggleButton.off_color),
    XtRString,
    "black"
  },
  {
    XcNonColor,
    XcCColor,
    XtRString,
    sizeof(String),
    offset(toggleButton.on_color),
    XtRString,
    "green"
  },
  {
    XcNstate,
    XcCState,
    XcRToggleState,
    sizeof(XcToggleState),
    offset(toggleButton.state),
    XtRString,
    "off"
  },
  {
    XcNactivateCallback,
    XtCCallback,
    XtRCallback,
    sizeof(caddr_t),
    offset(toggleButton.callback),
    XtRCallback,
    NULL
  },

};


/* Widget translation and actions tables. */
static XtActionsRec actions[] =
{
   { "Do_button", Do_button },
   { "Do_enter_leave", Do_enter_leave }
};

static char defaultTranslations[] = "<Btn1Down>:  Do_button() \n\
				     <Enter>:  Do_enter_leave() \n\
				     <Leave>:  Do_enter_leave() \n\
				     <Btn1Up>:  Do_button() ";


/* Widget Class Record initialization */
ToggleButtonClassRec toggleButtonClassRec =
{
  {
  /* core_class part */
    (WidgetClass) &controlClassRec,		/* superclass */
    "ToggleButton",				/* class_name */
    sizeof(ToggleButtonRec),			/* widget_size */
    ClassInitialize,				/* class_initialize */
    NULL,					/* class_part_initialize */
    FALSE,					/* class_inited */
    Initialize,					/* initialize */
    NULL,					/* initialize_hook */
    XtInheritRealize,				/* realize */
    actions,					/* actions */
    XtNumber(actions),				/* num_actions */
    resources,					/* resources */
    XtNumber(resources),			/* num_resources */
    NULLQUARK,					/* xrm_class */
    TRUE,					/* compress_motion */
    TRUE,					/* compress_exposure */
    TRUE,					/* compress_enterleave */
    FALSE,					/* visible_interest */
    NULL,					/* destroy */
    Resize,					/* resize */
    Redisplay,					/* expose */
    SetValues,					/* set_values */
    NULL,					/* set_values_hook */
    XtInheritSetValuesAlmost,			/* set_values_almost */
    NULL,					/* get_values_hook */
    NULL,					/* accept_focus */
    XtVersion,					/* version */
    NULL,					/* callback_private */
    defaultTranslations,			/* tm_table */
    QueryGeometry,				/* query_geometry */
    XtInheritDisplayAccelerator,		/* display_accelerator */
    NULL					/* extension */
  },
  {
  /* Control class part */
    0,						/* dummy_field */
  },
  {
  /* ToggleButton class part */
    0,						/* dummy_field */
  },
};

WidgetClass xcToggleButtonWidgetClass = (WidgetClass)&toggleButtonClassRec;


/* Widget method function definitions */

/*******************************************************************
 NAME:		ClassInitialize.	
 DESCRIPTION:
   This method initializes the ToggleButton widget class. Specifically,
it registers resource value converter functions with Xt.

*******************************************************************/

static void ClassInitialize()
{

   XtAddConverter(XtRString, XcRToggleState, CvtStringToToggleState, NULL, 0);

}  /* end of ClassInitialize */



/*******************************************************************
 NAME:		CvtStringToToggleState.		
 DESCRIPTION:
   This function converts resource settings in string form to the
XcRToggleState representation type.

*******************************************************************/

void CvtStringToToggleState(args, num_args, fromVal, toVal)
XrmValuePtr args;		/* unused */
Cardinal *num_args;		/* unused */
XrmValuePtr fromVal;
XrmValuePtr toVal;
{
/* Local variables */
static XcToggleState state;
char lowerstring[100];

/* Convert the resource string to lower case for quick comparison */
   ToLower((char *)fromVal->addr, lowerstring);

/*
 * Compare resource string with valid XcToggleState strings and assign to
 * datatype.
 */
   if (strcmp(lowerstring, XcEon) == 0) 
   {
      state = XcOn;
      CvtDone(XcToggleState, &state);
   }
   else if (strcmp(lowerstring, XcEoff) == 0) 
   {
      state = XcOff;
      CvtDone(XcToggleState, &state);
   }

/*
 * If the string is not valid for this resource type, print a warning
 * and do not make the conversion.
 */
   XtStringConversionWarning(fromVal->addr, "XcToggleState");
   toVal->addr = NULL;
   toVal->size = 0;

}  /* end of CvtStringToToggleState */




/*******************************************************************
 NAME:		Initialize.		
 DESCRIPTION:
   This is the initialize method for the ToggleButton widget.  It 
validates user-modifiable instance resources and initializes private 
widget variables and structures.  This function also creates any server 
resources (i.e., GCs, fonts, Pixmaps, etc.) used by this widget.  This
method is called by Xt when the application calls XtCreateWidget().

*******************************************************************/

static void Initialize(request, new)
ToggleButtonWidget request, new;
{
/* Local variables */
Display *display = XtDisplay(new);
int scr = DefaultScreen(XtDisplay(new));
Colormap cmap = DefaultColormap(display, scr);
XColor color;
char errstring[40];

DPRINTF(("ToggleButton: executing Initialize...\n"));

/*
 * Validate public instance variable settings.
 */
  if ((new->toggleButton.state != XcOn) && (new->toggleButton.state != XcOff))  
  {
     sprintf(errstring, "ToggleButton: invalid Toggle state: %d", 
						new->toggleButton.state);
     XtWarning(errstring);
     new->toggleButton.state = XcOff;
  }  


/* Initialize the ToggleButton width and height. */
   if (new->core.width < MIN_TB_WIDTH)
      new->core.width = MIN_TB_WIDTH; 
   if (new->core.height < MIN_TB_HEIGHT)
      new->core.height = MIN_TB_HEIGHT; 

/* Initialize private instance variables.  */
   new->toggleButton.armed = False;

/* Set the initial geometry of the ToggleButton elements. */
   Resize(new);


/* convert on/off colors to its equivalent color pixel */
   if (XParseColor(display, cmap, new->toggleButton.on_color, &color))
      if (XAllocColor(display, cmap, &color))
         new->toggleButton.on_pixel = color.pixel;
      else
      {
         sprintf(errstring, "ToggleButton: Couldn't allocate color: %s", 
		 new->toggleButton.on_color);
         XtWarning(errstring);
         strcpy(new->toggleButton.on_color, "white");
         new->toggleButton.on_pixel = WhitePixel(display, scr);
      }
   else
   {
      sprintf(errstring, "ToggleButton: Couldn't parse color: %s", 
		 new->toggleButton.on_color);
      XtWarning(errstring);
      strcpy(new->toggleButton.on_color, "white");
      new->toggleButton.on_pixel = WhitePixel(display, scr); }

   if (XParseColor(display, cmap, new->toggleButton.off_color, &color))
      if (XAllocColor(display, cmap, &color))
         new->toggleButton.off_pixel = color.pixel;
      else
      {
         sprintf(errstring, "ToggleButton: Couldn't allocate color: %s", 
		 new->toggleButton.off_color);
         XtWarning(errstring);
         strcpy(new->toggleButton.off_color, "black");
         new->toggleButton.on_pixel = BlackPixel(display, scr);
      }
   else
   {
      sprintf(errstring, "ToggleButton: Couldn't parse color: %s", 
		 new->toggleButton.off_color);
      XtWarning(errstring);
      strcpy(new->toggleButton.off_color, "black");
      new->toggleButton.on_pixel = BlackPixel(display, scr);
   }

DPRINTF(("ToggleButton: done Initialize\n"));
DPRINTF(("ToggleButton: done Initialize\n"));

}  /* end of Initialize */





/*******************************************************************
 NAME:		Redisplay.	
 DESCRIPTION:
   This function is the ToggleButton's Expose method.  It redraws the
ToggleButton's 3D rectangle background, and label string. All drawing 
takes place within the widget's window (no need for an off-screen pixmap).

*******************************************************************/

static void Redisplay(w, event)
ToggleButtonWidget w;
XExposeEvent *event;
{

/*
 * Check to see whether or not the widget's window is mapped.  You can't
 * draw into a window that is not mapped.  Realizing a widget doesn't 
 * mean its mapped, but this call will work for most Window Managers.
 */
   if (!XtIsRealized(w))
      return;

DPRINTF(("ToggleButton: executing Redisplay\n"));

   Draw_button(w, w->control.gc, RAISED);
DPRINTF(("ToggleButton: done Redisplay\n"));

}  /* end of Redisplay */





/*******************************************************************
 NAME:		SetValues.
 DESCRIPTION:
   This is the set_values method for this widget. It validates resource
settings set with XtSetValues. If a resource is changed that would
require re-drawing the widget, return True.

*******************************************************************/

static Boolean SetValues(cur, req, new)
ToggleButtonWidget cur, req, new;
{
/* Local variables */
Boolean do_redisplay = False;

DPRINTF(("ToggleButton: executing SetValues for %ld\n", cur));

/* Validate new resource settings. */

/* Check if the label has changed */
   if (strcmp(new->control.label, cur->control.label) != 0)
   {
      do_redisplay = True;
      Resize(new);
   }


DPRINTF(("ToggleButton: done SetValues\n"));
   return do_redisplay;


}  /* end of SetValues */




/*******************************************************************
 NAME:		Resize.
 DESCRIPTION:
   This is the resize method of the ToggleButton widget. It resizes the
ToggleButton's graphics based on the new width and height of the widget's
window.  

*******************************************************************/

static void Resize(w)
ToggleButtonWidget w;
{
  Pixel facewidth2;
  Pixel textwidth2;

  DPRINTF(("ToggleButton: executing Resize\n"));

/* Set the widgets new width and height. */
   w->toggleButton.face.x = w->toggleButton.face.y = w->control.shade_depth;
   w->toggleButton.face.width = w->core.width - (2*w->control.shade_depth)+1;
   w->toggleButton.face.height = w->core.height - (2*w->control.shade_depth)+1;


/* Set the new label location.  Center it in the button. */
   facewidth2 = w->toggleButton.face.width / 2;
   textwidth2 = XTextWidth(w->control.font, w->control.label,
			   strlen(w->control.label)) / 2;
   textwidth2 = (textwidth2 < facewidth2) ? textwidth2 : facewidth2; 

   w->toggleButton.lbl.x = w->toggleButton.face.x + facewidth2 - textwidth2;
   w->toggleButton.lbl.y =  (w->toggleButton.face.height / 4) + 
		 	    ((w->control.font->ascent + 
			      w->control.font->descent) / 2);


/* Set the new indicator location.  Bottom half center */
   w->toggleButton.indicator.x = w->toggleButton.face.x + 
				 w->toggleButton.face.width / 6;
   w->toggleButton.indicator.y = 2 * (w->toggleButton.face.height / 3);
   w->toggleButton.indicator.width = 2 * (w->toggleButton.face.width / 3); 
   w->toggleButton.indicator.height = (w->toggleButton.face.y + 
					w->toggleButton.face.height) -
						w->toggleButton.indicator.y;
   DPRINTF(("ToggleButton: done Resize\n"));

}  /* end of Resize */




/*******************************************************************
 NAME:		QueryGeometry.		
 DESCRIPTION:
   This function is the widget's query_geometry method.  It simply
checks the proposed size and returns the appropriate value based on
the proposed size.  If the proposed size is greater than the maximum
appropriate size for this widget, QueryGeometry returns the recommended
size.

*******************************************************************/

static XtGeometryResult QueryGeometry(w, proposed, answer)
ToggleButtonWidget w;
XtWidgetGeometry *proposed, *answer;
{
/* Set the request mode mask for the returned answer. */
   answer->request_mode = CWWidth | CWHeight;

/* Set the recommended size. */
   answer->width = (w->core.width > MAX_TB_WIDTH)
	? MAX_TB_WIDTH : w->core.width;
   answer->height = (w->core.height > MAX_TB_HEIGHT)
	? MAX_TB_HEIGHT : w->core.height;

/* 
 * Check the proposed dimensions. If the proposed size is larger than
 * appropriate, return the recommended size.
 */
   if (((proposed->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight))
	&& proposed->width == answer->width 
	&& proposed->height == answer->height)
      return XtGeometryYes;
   else if (answer->width == w->core.width && answer->height == w->core.height)
      return XtGeometryNo;
   else
      return XtGeometryAlmost;

}  /* end of QueryGeometry */






/* Widget action functions. */

/*******************************************************************
 NAME:		Do_button.		
 DESCRIPTION:
   This function activates the OK Button and calls the application's
registered callback function.

*******************************************************************/

static void Do_button(w, event)
ToggleButtonWidget w;
XEvent *event;
{
/* Local variables */

/*
 * ButtonRelease event. Call the application's callback function
 * and pass the value's data type and new value.
 */
   if ((w->toggleButton.armed) && (event->type == ButtonRelease))
   {
      Draw_button(w, w->control.gc, RAISED);
      w->toggleButton.armed = False;

      if (w->toggleButton.state == XcOn)
         w->toggleButton.state = XcOff;
      else
         w->toggleButton.state = XcOn;

      Draw_ind_color(w, XtDisplay(w), XtWindow(w), w->control.gc);
   /* Call the callback functions. */
      XtCallCallbacks(w, XcNactivateCallback, NULL); 
   }
/*
 * ButtonPress event. Just depress the button and arm it.
 */
   else if (event->type == ButtonPress)
   {
      w->toggleButton.armed = True;
      Draw_button(w, w->control.gc, DEPRESSED);
   }

}  /* end of Do_button */




/*******************************************************************
 NAME:		Do_enter_leave.		
 DESCRIPTION:
   This function handles Enter and Leave Notify events.

*******************************************************************/

static void Do_enter_leave(w, event)
ToggleButtonWidget w;
XEvent *event;
{
/* Local variables */

/*
 * LeaveNotify event. 
 */
   if ((w->toggleButton.armed) && (event->type == LeaveNotify))
   {
      w->toggleButton.armed = False;
      Draw_button(w, w->control.gc, RAISED);
   }
/*
 * EnterNotify event.
 */
   else if ((event->type == EnterNotify) && 
			(event->xcrossing.state & Button1Mask))
   {
      w->toggleButton.armed = True;
      Draw_button(w, w->control.gc, DEPRESSED);
   }

}  /* end of Do_enter_leave */



/*******************************************************************
 NAME:		Draw_button.		
 DESCRIPTION:
   Redraws the button in the given state, RAISED or DEPRESSED.

*******************************************************************/

static void Draw_button(w, gc, type)
ToggleButtonWidget w;
GC gc;
Type3d type;
{

/* Draw the background button */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	 0, 0, w->core.width, w->core.height, type);

/* Draw the Label string. */
   XSetClipRectangles(XtDisplay(w), w->control.gc, 0, 0, 
  			&(w->toggleButton.face), 1, Unsorted); 
   XSetForeground(XtDisplay(w), w->control.gc,
                        w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), w->control.gc,
	w->toggleButton.lbl.x, w->toggleButton.lbl.y, 
	w->control.label, strlen(w->control.label));

/* Draw the 3D rectangle background for the indicator in the given type. */
   XSetClipMask(XtDisplay(w), gc, None);
   Rect3d(w, XtDisplay(w), XtWindow(w), gc,
                w->toggleButton.indicator.x,
		w->toggleButton.indicator.y,
		w->toggleButton.indicator.width,
		w->toggleButton.indicator.height,
		DEPRESSED);
   Draw_ind_color(w, XtDisplay(w), XtWindow(w), gc);
}  /* end of Draw_button */

void Draw_ind_color(w, display, drawable, gc)
ToggleButtonWidget w;
Display *display;
Drawable drawable;
GC gc;
{

/* Clear indicator box by re-filling it with is background color */
   if (w->toggleButton.state == XcOn)
      XSetForeground(display, gc, w->toggleButton.on_pixel);
   else 
      XSetForeground(display, gc, w->toggleButton.off_pixel);

   XFillRectangle(display, drawable, gc, 
	w->toggleButton.indicator.x + w->control.shade_depth,
        w->toggleButton.indicator.y + w->control.shade_depth,
        w->toggleButton.indicator.width - (2 * w->control.shade_depth),
        w->toggleButton.indicator.height - (2 * w->control.shade_depth));
} /* end of Draw_ind_color */

void upstring(str)
char *str;
{
  short i = 0;
  
  while (str[i])
  {
    str[i] = toupper(str[i]);
    i++;
  }
} /* end of upstring */

/**************************************************************
 NAME:		XcToggleSet
 DESCRIPTION:
	This Function sets toggle state on or off.
***************************************************************/

void XcToggleSet(w, state)
ToggleButtonWidget w;
XcToggleState state;
{
  char errstring[40];

  if ((state != XcOn) && (state != XcOff))  
  {
     sprintf(errstring, "ToggleButton: Unknown Toggle state: %d", state);
     XtWarning(errstring);
     w->toggleButton.state = XcOff;
  }  
  else
     w->toggleButton.state = state;

  Draw_ind_color(w, XtDisplay(w), XtWindow(w), w->control.gc);

} /* end of XcToggleSet */



/**************************************************************
 NAME:		XcToggleGet
 DESCRIPTION:
	This Function gets toggle state value.
***************************************************************/

void XcToggleGet(w, state)
ToggleButtonWidget w;
XcToggleState *state;
{

     *state = w->toggleButton.state;
  
} /* end of XcToggleGet */


/* end of ToggleButton.c */

