/*******************************************************************
 FILE:		ValueButton.c
 CONTENTS:	Definitions for structures, methods, and actions of the 
		ValueButton widget.
 AUTHOR:	Paul D. Johnston
 HISTORY:
 Date		Action
 ---------	------------------------------------
 06/14/92	Added the ability to increment/decrement value by using
		the up/down arrow keys or the +/- keys.  The Accept Button
		action is duplicated with the RETURN or ENTER keys.
 10/22/91	Created.

********************************************************************/

#include <stdio.h>

/* Xlib includes */
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

/* Xt includes */
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#ifdef X11R3
#include <X11/Xmu.h>
#else
#include <X11/Xmu/Xmu.h>
#endif


/* Widget includes */
#include <X11/Xc/Xc.h>
#include <X11/Xc/Control.h>
#include <X11/Xc/VButtonP.h>	/* (includes VButton.h also) */


/* Macro redefinition for offset. */
#define offset(field) XtOffset(ValueButtonWidget, field)


/* Declare widget methods */
static void ClassInitialize();
static void Initialize();
static void Redisplay();
static void Resize();
static XtGeometryResult QueryGeometry();
static Boolean SetValues();

/* Declare widget actions */
static void HandleBtnPress();	/* Handles all ButtonPress events
				 * in the ValueButton widget.
				 */
static void HandleBtnRelease();	/* Handles all ButtonRelease events
				 * in the ValueButton widget.
				 */
static void HandleKeyPress();	/* Handles all KeyPress events
				 * in the ValueButton widget.
				 */



/* Declare functions and variables private to this widget */
static void Draw_value();
static void Call_callback();
static void Do_accel();



/* Define the widget's resource list (None right now). */


/* Widget translation and actions tables. */
static XtActionsRec actions[] =
{
   { "HandleBtnPress", HandleBtnPress },
   { "HandleBtnRelease", HandleBtnRelease },
   { "HandleKeyPress", HandleKeyPress }
};

static char defaultTranslations[] = "<Btn1Down>:  HandleBtnPress() \n\
				     <Btn1Up>:  HandleBtnRelease() \n\
				     <Key>:  HandleKeyPress() ";


/* Widget Class Record initialization */
ValueButtonClassRec valueButtonClassRec =
{
  {
  /* core_class part */
    (WidgetClass) &valueClassRec,		/* superclass */
    "ValueButton",				/* class_name */
    sizeof(ValueButtonRec),			/* widget_size */
    ClassInitialize,				/* class_initialize */
    NULL,					/* class_part_initialize */
    FALSE,					/* class_inited */
    Initialize,					/* initialize */
    NULL,					/* initialize_hook */
    XtInheritRealize,				/* realize */
    actions,					/* actions */
    XtNumber(actions),				/* num_actions */
    NULL,					/* resources */
    0,						/* num_resources */
    NULLQUARK,					/* xrm_class */
    TRUE,					/* compress_motion */
    TRUE,					/* compress_exposure */
    TRUE,					/* compress_enterleave */
    TRUE,					/* visible_interest */
    NULL,					/* destroy */
    Resize,					/* resize */
    Redisplay,					/* expose */
    SetValues,					/* set_values */
    NULL,					/* set_values_hook */
    XtInheritSetValuesAlmost,			/* set_values_almost */
    NULL,					/* get_values_hook */
    NULL,					/* accept_focus */
    XtVersion,					/* version */
    NULL,					/* callback_private */
    defaultTranslations,			/* tm_table */
    QueryGeometry,				/* query_geometry */
    NULL,					/* display_accelerator */
    NULL,					/* extension */
  },
  {
  /* Control class part */
    0,						/* dummy_field */
  },
  {
  /* Value class part */
    0,						/* dummy_field */
  },
  {
  /* ValueButton class part */
    0,						/* dummy_field */
  }
};

WidgetClass xcValueButtonWidgetClass = (WidgetClass)&valueButtonClassRec;


/* Widget method function definitions */

/*******************************************************************
 NAME:		ClassInitialize.	
 DESCRIPTION:
   This method initializes the ValueButton widget class. Specifically,
it registers resource value converter functions with Xt.

*******************************************************************/

static void ClassInitialize()
{

   XtAddConverter(XtRString, XtROrientation, XmuCvtStringToOrientation,
						NULL, 0);

}  /* end of ClassInitialize */






/*******************************************************************
 NAME:		Initialize.		
 DESCRIPTION:
   This is the initialize method for the ValueButton widget.  It 
validates user-modifiable instance resources and initializes private 
widget variables and structures.  This function also creates any server 
resources (i.e., GCs, fonts, Pixmaps, etc.) used by this widget.  This
method is called by Xt when the application calls XtCreateWidget().

*******************************************************************/

static void Initialize(request, new)
ValueButtonWidget request, new;
{
/* Local variables */
Display *display = XtDisplay(new);
int scr = DefaultScreen(XtDisplay(new));


DPRINTF(("ValueButton: executing Initialize...\n"));

/*
 * Validate public instance variable settings.
 */
/* Initialize the ValueButton width and height. */
   if (new->core.width < MIN_VB_WIDTH)
      new->core.width = MIN_VB_WIDTH; 
   if (new->core.height < MIN_VB_HEIGHT)
      new->core.height = MIN_VB_HEIGHT; 
   
/* Initialize private instance variables.  */
   new->valueButton.acpt_armed = new->valueButton.up_armed = 
   				new->valueButton.down_armed = False;


/* Set the initial geometry of the ValueButton elements. */
   Resize(new);


DPRINTF(("ValueButton: done Initialize\n"));

}  /* end of Initialize */





/*******************************************************************
 NAME:		Redisplay.	
 DESCRIPTION:
   This function is the ValueButton's Expose method.  It redraws the
ValueButton's 3D rectangle background, Value Box, label and value string,
and the Up and Down Arrows.  All drawing takes place within the widget's 
window (no need for an off-screen pixmap).

*******************************************************************/

static void Redisplay(w, event)
ValueButtonWidget w;
XExposeEvent *event;
{

/*
 * Check to see whether or not the widget's window is mapped.  You can't
 * draw into a window that is not mapped.  Realizing a widget doesn't 
 * mean its mapped, but this call will work for most Window Managers.
 */
   if (!XtIsRealized(w))
      return;

DPRINTF(("ValueButton: executing Redisplay\n"));

/* Draw the 3D rectangle background for the ValueButton. */
   XSetClipMask(XtDisplay(w), w->control.gc, None);
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		0, 0, w->core.width, w->core.height, RAISED);

/* Draw the Accept Button */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->valueButton.acpt.x, w->valueButton.acpt.y,
	w->valueButton.acpt.width, w->valueButton.acpt.height, RAISED);

/* Draw the Value Box */
   Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->value.value_box.x - w->control.shade_depth, 
	w->value.value_box.y - w->control.shade_depth,
	w->value.value_box.width + (2 * w->control.shade_depth + 1),  
	w->value.value_box.height + (2 * w->control.shade_depth + 1), 
							DEPRESSED);

/* Draw the new value string in the Value Box. */
   Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);


/* Draw the Up and Down Arrows. */
   Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.up_arrow), UP, RAISED);
   Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.down_arrow), DOWN, RAISED);


/* Draw the Label string. */
   XSetClipRectangles(XtDisplay(w), w->control.gc, 0, 0, 
  			&(w->valueButton.face), 1, Unsorted); 
   XSetForeground(XtDisplay(w), w->control.gc,
                        w->control.label_pixel);
   XDrawString(XtDisplay(w), XtWindow(w), w->control.gc,
	w->valueButton.lbl.x, w->valueButton.lbl.y, 
	w->control.label, strlen(w->control.label));

DPRINTF(("ValueButton: done Redisplay\n"));

}  /* end of Redisplay */





/*******************************************************************
 NAME:		SetValues.
 DESCRIPTION:
   This is the set_values method for this widget. It validates resource
settings set with XtSetValues. If a resource is changed that would
require re-drawing the widget, return True.

*******************************************************************/

static Boolean SetValues(cur, req, new)
ValueButtonWidget cur, req, new;
{
/* Local variables */
Boolean do_redisplay = False;


DPRINTF(("ValueButton: executing SetValues \n"));

/* Validate new resource settings. */


DPRINTF(("ValueButton: done SetValues\n"));
   return do_redisplay;


}  /* end of SetValues */




/*******************************************************************
 NAME:		Resize.
 DESCRIPTION:
   This is the resize method of the ValueButton widget. It resizes the
ValueButton's graphics based on the new width and height of the widget's
window.  

*******************************************************************/

static void Resize(w)
ValueButtonWidget w;
{

DPRINTF(("ValueButton: executing Resize\n"));

/* Set the widgets new width and height. */
   w->valueButton.face.x = w->valueButton.face.y = w->control.shade_depth;
   w->valueButton.face.width = w->core.width - (2*w->control.shade_depth)+1;
   w->valueButton.face.height = w->core.height - (2*w->control.shade_depth)+1;

/* Establish the new Value Box geometry. */
   w->value.value_box.x = (short)(w->valueButton.face.width / 2 + 0.5)
					+ w->control.shade_depth;
   w->value.value_box.y = (short)(w->valueButton.face.height / 2 + 0.5)
					+ w->control.shade_depth;
   w->value.value_box.width = w->valueButton.face.width -
                        (w->value.value_box.x + w->control.shade_depth) - 
						(2*w->control.shade_depth);
   w->value.value_box.height = w->value.value_box.y -
						(2*w->control.shade_depth);

/* Establish the position of the value string within the Value Box. */
   Position_val(w);

/* Set the new label location. */
   w->valueButton.lbl.x = w->value.value_box.x - w->control.shade_depth;
   w->valueButton.lbl.y = w->valueButton.face.y + 
		((w->value.value_box.y - w->control.shade_depth - 
		  		w->valueButton.face.y) / 2 + 0.5) +
		((w->control.font->ascent + 
			w->control.font->descent) / 2)
		- w->control.font->descent; 


/* Resize the Up and Down Arrows. */
   w->valueButton.up_arrow.x = w->valueButton.down_arrow.x =
				(w->valueButton.face.width / 8 + 0.5);
   w->valueButton.up_arrow.y = (w->valueButton.face.height / 16 + 0.5);
   w->valueButton.down_arrow.y = (10 * w->valueButton.face.height / 16 + 0.5);
   w->valueButton.up_arrow.width = w->valueButton.down_arrow.width =
				(w->valueButton.face.width /4 + 0.5);
   w->valueButton.up_arrow.height = w->valueButton.down_arrow.height =
				(5 * w->valueButton.face.height / 16 + 0.5);


/* Establish the new Accept Button geometry. */
   w->valueButton.acpt.x = w->valueButton.up_arrow.x; 
   w->valueButton.acpt.y = (7 * w->valueButton.face.height / 16 + 0.5); 
   w->valueButton.acpt.width = w->valueButton.up_arrow.width;
   w->valueButton.acpt.height = (2 * w->valueButton.face.height / 16 + 0.5); 

DPRINTF(("ValueButton: done Resize\n"));

}  /* end of Resize */




/*******************************************************************
 NAME:		QueryGeometry.		
 DESCRIPTION:
   This function is the widget's query_geometry method.  It simply
checks the proposed size and returns the appropriate value based on
the proposed size.  If the proposed size is greater than the maximum
appropriate size for this widget, QueryGeometry returns the recommended
size.

*******************************************************************/

static XtGeometryResult QueryGeometry(w, proposed, answer)
ValueButtonWidget w;
XtWidgetGeometry *proposed, *answer;
{
/* Set the request mode mask for the returned answer. */
   answer->request_mode = CWWidth | CWHeight;

/* Set the recommended size. */
   answer->width = (w->core.width > MAX_VB_WIDTH)
	? MAX_VB_WIDTH : w->core.width;
   answer->height = (w->core.height > MAX_VB_HEIGHT)
	? MAX_VB_HEIGHT : w->core.height;

/* 
 * Check the proposed dimensions. If the proposed size is larger than
 * appropriate, return the recommended size.
 */
   if (((proposed->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight))
	&& proposed->width == answer->width 
	&& proposed->height == answer->height)
      return XtGeometryYes;
   else if (answer->width == w->core.width && answer->height == w->core.height)
      return XtGeometryNo;
   else
      return XtGeometryAlmost;

}  /* end of QueryGeometry */





/* Widget action functions. */

/*******************************************************************
 NAME:		HandleBtnPress.
 DESCRIPTION:
   This function manages the value manipulated with the ValueButton,
and handles graphic indications that an arrow or button was pressed 
to the user.

*******************************************************************/

static void HandleBtnPress(w, event)
ValueButtonWidget w;
XButtonEvent *event;
{

/*
 * Check to see if the pointer button was pressed within an arrow.
 * If it was, press the arrow and set its armed flag. 
 */
   if (Point_In_Rect(event->x, event->y, w->valueButton.up_arrow.x, 
			w->valueButton.up_arrow.y, 
			w->valueButton.up_arrow.width,
			w->valueButton.up_arrow.height))
   {
   /* Highlight the arrow. */
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.up_arrow), UP, DEPRESSED);

      w->valueButton.up_armed = True;
      w->valueButton.interval_id = XtAppAddTimeOut(
		XtWidgetToApplicationContext(w), 1000, Do_accel, w);
   }
   else if (Point_In_Rect(event->x, event->y, w->valueButton.down_arrow.x, 
			w->valueButton.down_arrow.y, 
			w->valueButton.down_arrow.width, 
			w->valueButton.down_arrow.height))
   {
   /* Highlight the arrow. */
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.down_arrow), DOWN, DEPRESSED);

      w->valueButton.down_armed = True;
      w->valueButton.interval_id = XtAppAddTimeOut(
		XtWidgetToApplicationContext(w), 1000, Do_accel, w);
   }
/*
 * ButtonPress on the Accept Button. Depress the button.
 */
   else if (Point_In_Rect(event->x, event->y, w->valueButton.acpt.x, 
			w->valueButton.acpt.y, 
			w->valueButton.acpt.width, 
			w->valueButton.acpt.height))
   {
   /* Depress the Accept Button */
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->valueButton.acpt.x, w->valueButton.acpt.y,
	w->valueButton.acpt.width, w->valueButton.acpt.height, DEPRESSED);

      w->valueButton.acpt_armed = True;
   }


}  /* end of HandleBtnPress */




/*******************************************************************
 NAME:		HandleBtnRelease.		
 DESCRIPTION:
   This function activates the Accept Button and calls the application's
registered callback function.

*******************************************************************/

static void HandleBtnRelease(w, event)
ValueButtonWidget w;
XButtonEvent *event;
{
/* Local variables */
static XcCallData call_data;
Boolean value_changed = False;

/*
 * If the pointer button was released within an arrow, raise the arrow, 
 * increment/decrement the value, and display the new value.
 */
   if (w->valueButton.up_armed)
   {
      if (Point_In_Rect(event->x, event->y, w->valueButton.up_arrow.x, 
			w->valueButton.up_arrow.y, 
			w->valueButton.up_arrow.width,
			w->valueButton.up_arrow.height))
      {
      /* Increment the value based on the value's data type */
         if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
         {
	    w->value.val.lval += w->value.increment.lval;
	    if (w->value.val.lval > w->value.upper_bound.lval)
	       w->value.val.lval = w->value.upper_bound.lval;
            else
	       value_changed = True;
         }
         else if (w->value.datatype == XcFval) 
         {
	    w->value.val.fval += w->value.increment.fval;
	    if (w->value.val.fval > w->value.upper_bound.fval)
	       w->value.val.fval = w->value.upper_bound.fval;
            else
	       value_changed = True;
         }
      }

   /* Un-highlight the arrow. */
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.up_arrow), UP, RAISED);
      w->valueButton.up_armed = False;
   }
   else if (w->valueButton.down_armed)
   {
      if (Point_In_Rect(event->x, event->y, w->valueButton.down_arrow.x, 
			w->valueButton.down_arrow.y, 
			w->valueButton.down_arrow.width, 
			w->valueButton.down_arrow.height))
      {
      /* Decrement the value based on the value's data type */
         if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
         {
	    w->value.val.lval -= w->value.increment.lval;
	    if (w->value.val.lval < w->value.lower_bound.lval)
	       w->value.val.lval = w->value.lower_bound.lval;
            else
	       value_changed = True;
         }
         else if (w->value.datatype == XcFval)
         {
	    w->value.val.fval -= w->value.increment.fval;
	    if (w->value.val.fval < w->value.lower_bound.fval)
	       w->value.val.fval = w->value.lower_bound.fval;
            else
	       value_changed = True;
         }
      }

   /* Un-highlight the arrow. */
      Arrow3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
		&(w->valueButton.down_arrow), DOWN, RAISED);
      w->valueButton.down_armed = False;
   }
/*
 * ButtonRelease in the Accept Button. Call the application's callback 
 * function.
 */
   else if (w->valueButton.acpt_armed)
   {
   /* Raise the Accept Button */
      Rect3d(w, XtDisplay(w), XtWindow(w), w->control.gc,
	w->valueButton.acpt.x, w->valueButton.acpt.y,
	w->valueButton.acpt.width, w->valueButton.acpt.height, RAISED);

      w->valueButton.acpt_armed = False;

      if (Point_In_Rect(event->x, event->y, w->valueButton.acpt.x, 
			w->valueButton.acpt.y, 
			w->valueButton.acpt.width, 
			w->valueButton.acpt.height))
	 Call_callback(w);
   }

/* Update the value if it has changed. */
   if (value_changed)
      Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);

}  /* end of HandleBtnRelease */




/*******************************************************************
 NAME:		HandleKeyPress.		
 DESCRIPTION:
   This function increments/decrements the ValueButton's value based on
key presses of the up/down arrow or +/- keys.  It also calls the
application's callback function(s) if the RETURN/ENTER key was pressed.

*******************************************************************/

static void HandleKeyPress(w, event)
ValueButtonWidget w;
XKeyEvent *event;
{
/* Local variables */
KeySym key;
Boolean value_changed = False;

/* Only process Up and Down arrow key presses */
   key = XLookupKeysym(event, 0);
   if ((key == XK_Up) || (key == XK_KP_Add) || (key == XK_plus))   
   {
   /* Increment the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))   
      {
	 w->value.val.lval += w->value.increment.lval;
	 if (w->value.val.lval > w->value.upper_bound.lval)
	    w->value.val.lval = w->value.upper_bound.lval;
         else
	    value_changed = True;
      }
      else if (w->value.datatype == XcFval)
      {
	 w->value.val.fval += w->value.increment.fval;
	 if (w->value.val.fval > w->value.upper_bound.fval)
	    w->value.val.fval = w->value.upper_bound.fval;
         else
	    value_changed = True;
      }
   }
   else if ((key == XK_Down) || (key == XK_KP_Subtract) || (key == XK_minus))
   {
   /* Decrement the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))   
      {
	 w->value.val.lval -= w->value.increment.lval;
	 if (w->value.val.lval < w->value.lower_bound.lval)
	    w->value.val.lval = w->value.lower_bound.lval;
         else
	    value_changed = True;
      }
      else if (w->value.datatype == XcFval)
      {
	 w->value.val.fval -= w->value.increment.fval;
	 if (w->value.val.fval < w->value.lower_bound.fval)
	    w->value.val.fval = w->value.lower_bound.fval;
         else
	    value_changed = True;
      }
   }
   else if ((key == XK_Return) || (key == XK_KP_Enter))
   {
      Call_callback(w);
   }

/* Update the displayed value accordingly. */
   if (value_changed)
   {
      Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);
   }

}  /* end of HandleKeyPress */




/*******************************************************************
 NAME:		Draw_value.		
 DESCRIPTION:
   This function redraws the value string in the Value Box.

*******************************************************************/

static void Draw_value(w, display, drawable, gc)
ValueButtonWidget w;
Display *display;
Drawable drawable;
GC gc;
{
/* Local variables */
char *temp;
XRectangle val_clip;

/* Clear the Value Box by re-filling it with its background color. */
   XSetForeground(display, gc, w->value.value_bg_pixel); 
   XFillRectangle(display, drawable, gc,
  	w->value.value_box.x, w->value.value_box.y, 
  	w->value.value_box.width, w->value.value_box.height); 

/*
 * Now draw the value string in its foreground color, clipped by the
 * Value Box.
 */
   XSetForeground(display, gc, w->value.value_fg_pixel); 
   XSetClipRectangles(display, gc, 0, 0, 
  	&(w->value.value_box), 1, Unsorted); 

   temp = Print_value(w->value.datatype, &w->value.val, w->value.decimals);

   Position_val(w);

   XDrawString(display, drawable, gc,
    	w->value.vp.x, w->value.vp.y, temp, strlen(temp)); 

/* Reset the clip_mask to no clipping. */
   XSetClipMask(display, gc, None);

}  /* end of Draw_value */





/*******************************************************************
 NAME:		Call_callback.		
 DESCRIPTION:
   This function prepares the call_data structure and calls the
application's callback function with this data.

*******************************************************************/

static void Call_callback(w)
ValueButtonWidget w;
{
/* Local variables */
static XcCallData call_data;

/* Prepare the call_data */
   call_data.dtype = w->value.datatype;
   call_data.decimals = w->value.decimals;
   if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))  
      call_data.value.lval = w->value.val.lval;
   else if (w->value.datatype == XcFval)
      call_data.value.fval = w->value.val.fval;

/* Call the application's callback function(s). */
   if (w->value.callback != NULL)
      XtCallCallbacks(w, XcNvalueCallback, &call_data);

}  /* end of Call_callback */


/*******************************************************************
 NAME:		Do_accel.		
 DESCRIPTION:
   This function handles TimeOut events by incrementing/decrementing
the value if the pointer button is held down within an Arrow, thereby
implementing the acceleration feature.

*******************************************************************/

static void Do_accel(client_data, id)
caddr_t client_data;
XtIntervalId *id;		/* unused */
{
/* Local variables */
static XcCallData call_data;
ValueButtonWidget w = (ValueButtonWidget)client_data;
Boolean value_changed = False;
Window root, child;
int root_x, root_y, x, y;
unsigned int keys_buttons;

/* Get the current position of the pointer and button states. */
   if (XQueryPointer(XtDisplay(w), XtWindow(w), &root, &child,
		&root_x, &root_y,  &x, &y, &keys_buttons) == False)
      return;

/*
 * If the pointer button is still held within an arrow, increment/decrement 
 * the value, and display the new value.
 */
   if ((w->valueButton.up_armed) && (Point_In_Rect(x, y, 
		w->valueButton.up_arrow.x, w->valueButton.up_arrow.y,
				w->valueButton.up_arrow.width, 
				w->valueButton.up_arrow.height)))
   {
   /* Increment the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
      {
	 w->value.val.lval += w->value.increment.lval;
	 if (w->value.val.lval > w->value.upper_bound.lval)
	    w->value.val.lval = w->value.upper_bound.lval;
	 else
	    value_changed = True;
      }
      else if (w->value.datatype == XcFval) 
      {
	 w->value.val.fval += w->value.increment.fval;
	 if (w->value.val.fval > w->value.upper_bound.fval)
	    w->value.val.fval = w->value.upper_bound.fval;
	 else
	    value_changed = True;
      }
   }
   else if ((w->valueButton.down_armed) && (Point_In_Rect(x, y, 
		w->valueButton.down_arrow.x, w->valueButton.down_arrow.y,
				w->valueButton.down_arrow.width, 
				w->valueButton.down_arrow.height)))
   {
   /* Increment the value based on the value's data type */
      if ((w->value.datatype == XcLval) || (w->value.datatype == XcHval))
      {
	 w->value.val.lval -= w->value.increment.lval;
	 if (w->value.val.lval < w->value.lower_bound.lval)
	    w->value.val.lval = w->value.lower_bound.lval;
	 else
	    value_changed = True;
      }
      else if (w->value.datatype == XcFval) 
      {
	 w->value.val.fval -= w->value.increment.fval;
	 if (w->value.val.fval < w->value.lower_bound.fval)
	    w->value.val.fval = w->value.lower_bound.fval;
	 else
	    value_changed = True;
      }
   }

   if (value_changed)
   {
      Draw_value(w, XtDisplay(w), XtWindow(w), w->control.gc);
      w->valueButton.interval_id = XtAppAddTimeOut(
		XtWidgetToApplicationContext(w), 100, Do_accel, w);
   }

}  /* end of Do_accel */



/* end of ValueButton.c */

